# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""WorkingTree implementation tests for bzr.

This test the conformance of all the workingtre variations to the expected API.
Specific tests for individual formats are in the tests/test_workingtree file
rather than in tests/per_workingtree/*.py.
"""

from breezy import branchbuilder, tests, transport, workingtree
from breezy.tests import per_controldir, test_server
from breezy.transport import memory


def make_scenarios(
    transport_server,
    transport_readonly_server,
    formats,
    remote_server=None,
    remote_readonly_server=None,
    remote_backing_server=None,
):
    result = []
    for workingtree_format in formats:
        result.append(
            (
                workingtree_format.__class__.__name__,
                make_scenario(
                    transport_server, transport_readonly_server, workingtree_format
                ),
            )
        )
    default_wt_format = workingtree.format_registry.get_default()
    if remote_server is None:
        remote_server = test_server.SmartTCPServer_for_testing
    if remote_readonly_server is None:
        remote_readonly_server = test_server.ReadonlySmartTCPServer_for_testing
    if remote_backing_server is None:
        remote_backing_server = memory.MemoryServer
    scenario = make_scenario(remote_server, remote_readonly_server, default_wt_format)
    scenario["repo_is_remote"] = True
    scenario["vfs_transport_factory"] = remote_backing_server
    result.append((default_wt_format.__class__.__name__ + ",remote", scenario))
    return result


def make_scenario(transport_server, transport_readonly_server, workingtree_format):
    return {
        "transport_server": transport_server,
        "transport_readonly_server": transport_readonly_server,
        "bzrdir_format": workingtree_format._matchingcontroldir,
        "workingtree_format": workingtree_format,
    }


def wt_scenarios():
    """Returns the scenarios for all registered working trees.

    This can used by plugins that want to define tests against these working
    trees.
    """
    scenarios = make_scenarios(
        tests.default_transport,
        # None here will cause a readonly decorator to be created
        # by the TestCaseWithTransport.get_readonly_transport method.
        None,
        workingtree.format_registry._get_all(),
    )
    return scenarios


class TestCaseWithWorkingTree(per_controldir.TestCaseWithControlDir):
    def make_branch_and_tree(self, relpath, format=None):
        made_control = self.make_controldir(relpath, format=format)
        made_control.create_repository()
        b = made_control.create_branch()
        if getattr(self, "repo_is_remote", False):
            # If the repo is remote, then we just create a local lightweight
            # checkout
            # XXX: This duplicates a lot of Branch.create_checkout, but we know
            #      we want a) lightweight, and b) a specific WT format. We also
            #      know that nothing should already exist, etc.
            t = transport.get_transport(relpath)
            t.ensure_base()
            bzrdir_format = self.workingtree_format.get_controldir_for_branch()
            wt_dir = bzrdir_format.initialize_on_transport(t)
            branch_ref = wt_dir.set_branch_reference(b)
            wt = wt_dir.create_workingtree(None, from_branch=branch_ref)
        else:
            wt = self.workingtree_format.initialize(made_control)
        return wt

    def make_branch_builder(self, relpath, format=None):
        if format is None:
            format = self.workingtree_format.get_controldir_for_branch()
        builder = branchbuilder.BranchBuilder(
            self.get_transport(relpath), format=format
        )
        return builder


def load_tests(loader, standard_tests, pattern):
    test_names = [
        "add_reference",
        "add",
        "annotate_iter",
        "basis_inventory",
        "basis_tree",
        "break_lock",
        "canonical_path",
        "changes_from",
        "check",
        "check_state",
        "content_filters",
        "commit",
        "eol_conversion",
        "executable",
        "flush",
        "get_file_mtime",
        "get_parent_ids",
        "inv",
        "is_control_filename",
        "is_ignored",
        "locking",
        "merge_from_branch",
        "mkdir",
        "move",
        "nested_specifics",
        "parents",
        "paths2ids",
        "pull",
        "put_file",
        "readonly",
        "read_working_inventory",
        "remove",
        "rename_one",
        "revision_tree",
        "set_root_id",
        "shelf_manager",
        "smart_add",
        "symlinks",
        "transform",
        "uncommit",
        "unversion",
        "views",
        "walkdirs",
        "workingtree",
    ]
    test_workingtree_implementations = [
        "breezy.tests.per_workingtree.test_" + name for name in test_names
    ]

    scenarios = wt_scenarios()

    # add the tests for the sub modules
    return tests.multiply_tests(
        loader.loadTestsFromModuleNames(test_workingtree_implementations),
        scenarios,
        standard_tests,
    )


class TestWtScenarios(tests.TestCase):
    def test_protect_wt_scenarios(self):
        # Just make sure we don't accidentally delete the helper again
        wt_scenarios()



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the test trees used by the per_tree tests."""

import shutil

from breezy import errors
from breezy.tests import TestNotApplicable, TestSkipped, features, per_tree
from breezy.tree import MissingNestedTree


class TestTreeShapes(per_tree.TestCaseWithTree):
    def test_empty_tree_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_no_content(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        self.assertEqual([""], list(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [("", tree.path2id(""))],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual([""], [path for path, node in tree.iter_entries_by_dir()])

    def test_abc_tree_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        self.assertEqual({"", "a", "b", "b/c"}, set(tree.all_versioned_paths()))
        self.assertTrue(tree.is_versioned("a"))
        self.assertTrue(tree.is_versioned("b"))
        self.assertTrue(tree.is_versioned("b/c"))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["", "a", "b", "b/c"]],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                ["", "a", "b", "b/c"],
                [path for path, node in tree.iter_entries_by_dir()],
            )

        self.assertEqualDiff(b"contents of a\n", tree.get_file_text("a"))
        self.assertFalse(tree.is_executable("b/c"))

    def test_abc_tree_content_2_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content_2(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        self.assertEqual({"", "a", "b", "b/c"}, set(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["", "a", "b", "b/c"]],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["b"]],
                [
                    (path, node.file_id)
                    for path, node in tree.iter_entries_by_dir(specific_files=["b"])
                ],
            )
        else:
            self.assertEqual(
                ["", "a", "b", "b/c"],
                [path for path, node in tree.iter_entries_by_dir()],
            )
            self.assertEqual(
                ["b"],
                [path for path, node in tree.iter_entries_by_dir(specific_files=["b"])],
            )
        self.assertEqualDiff(b"foobar\n", tree.get_file_text("a"))
        self.assertFalse(tree.is_executable("b//c"))
        self.assertFalse(tree.is_executable("b/c"))

    def test_abc_tree_content_3_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content_3(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        self.assertEqual({"", "a", "b", "b/c"}, set(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["", "a", "b", "b/c"]],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                ["", "a", "b", "b/c"],
                [path for path, node in tree.iter_entries_by_dir()],
            )
        self.assertEqualDiff(b"contents of a\n", tree.get_file_text("a"))
        self.assertTrue(tree.is_executable("b/c"))

    def test_abc_tree_content_4_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content_4(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        self.assertEqual({"", "b", "d", "b/c"}, set(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["", "b", "d", "b/c"]],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                ["", "b", "d", "b/c"],
                [path for path, node in tree.iter_entries_by_dir()],
            )
        self.assertEqualDiff(b"contents of a\n", tree.get_file_text("d"))
        self.assertFalse(tree.is_executable("b/c"))

    def test_abc_tree_content_5_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content_5(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        self.assertEqual({"", "d", "b", "b/c"}, set(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in ["", "b", "d", "b/c"]],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                ["", "b", "d", "b/c"],
                [path for path, node in tree.iter_entries_by_dir()],
            )
        self.assertEqualDiff(b"bar\n", tree.get_file_text("d"))
        self.assertFalse(tree.is_executable("b/c"))

    def test_abc_tree_content_6_no_parents(self):
        tree = self.make_branch_and_tree(".")
        tree = self.get_tree_no_parents_abc_content_6(tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        expected_paths = (
            ["", "a"] + (["b"] if tree.has_versioned_directories() else []) + ["e"]
        )
        self.assertEqual(set(expected_paths), set(tree.all_versioned_paths()))
        if tree.supports_file_ids:
            self.assertEqual(
                [(p, tree.path2id(p)) for p in expected_paths],
                [(path, node.file_id) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                expected_paths, [path for path, node in tree.iter_entries_by_dir()]
            )
        self.assertEqualDiff(b"contents of a\n", tree.get_file_text("a"))
        self.assertTrue(tree.is_executable("e"))

    def test_tree_with_subdirs_and_all_content_types(self):
        # currently this test tree requires unicode. It might be good
        # to have it simply stop having the single unicode file in it
        # when dealing with a non-unicode filesystem.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.get_tree_with_subdirs_and_all_content_types()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        try:
            all_file_ids = set(tree.all_file_ids())
            tree.path2id("")
        except AttributeError:
            # doesn't support file ids
            all_file_ids = None
        if tree.has_versioned_directories():
            if all_file_ids is not None:
                self.assertEqual(
                    {
                        tree.path2id(p)
                        for p in [
                            "",
                            "0file",
                            "1top-dir",
                            "1top-dir/1dir-in-1topdir",
                            "1top-dir/0file-in-1topdir",
                            "symlink",
                            "2utf\u1234file",
                        ]
                    },
                    set(tree.all_file_ids()),
                )
            # note that the order of the paths and fileids is deliberately
            # mismatched to ensure that the result order is path based.
            self.assertEqual(
                [
                    ("", "directory"),
                    ("0file", "file"),
                    ("1top-dir", "directory"),
                    ("2utf\u1234file", "file"),
                    ("symlink", "symlink"),
                    ("1top-dir/0file-in-1topdir", "file"),
                    ("1top-dir/1dir-in-1topdir", "directory"),
                ],
                [(path, node.kind) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            if all_file_ids is not None:
                self.assertEqual(
                    {
                        tree.path2id(p)
                        for p in [
                            "",
                            "0file",
                            "1top-dir",
                            "1top-dir/0file-in-1topdir",
                            "symlink",
                            "2utf\u1234file",
                        ]
                    },
                    set(tree.all_file_ids()),
                )
            # note that the order of the paths and fileids is deliberately
            # mismatched to ensure that the result order is path based.
            self.assertEqual(
                [
                    ("", "directory"),
                    ("0file", "file"),
                    ("1top-dir", "directory"),
                    ("2utf\u1234file", "file"),
                    ("symlink", "symlink"),
                    ("1top-dir/0file-in-1topdir", "file"),
                ],
                [(path, node.kind) for path, node in tree.iter_entries_by_dir()],
            )

    def test_tree_with_subdirs_and_all_content_types_wo_symlinks(self):
        # currently this test tree requires unicode. It might be good
        # to have it simply stop having the single unicode file in it
        # when dealing with a non-unicode filesystem.
        tree = self.get_tree_with_subdirs_and_all_supported_content_types(False)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], tree.get_parent_ids())
        self.assertEqual([], tree.conflicts())
        self.assertEqual([], list(tree.unknowns()))
        # __iter__ has no strongly defined order
        if tree.has_versioned_directories():
            self.assertEqual(
                {
                    "",
                    "0file",
                    "1top-dir",
                    "1top-dir/0file-in-1topdir",
                    "1top-dir/1dir-in-1topdir",
                    "2utf\u1234file",
                },
                set(tree.all_versioned_paths()),
            )
            # note that the order of the paths and fileids is deliberately
            # mismatched to ensure that the result order is path based.
            self.assertEqual(
                [
                    ("", "directory"),
                    ("0file", "file"),
                    ("1top-dir", "directory"),
                    ("2utf\u1234file", "file"),
                    ("1top-dir/0file-in-1topdir", "file"),
                    ("1top-dir/1dir-in-1topdir", "directory"),
                ],
                [(path, node.kind) for path, node in tree.iter_entries_by_dir()],
            )
        else:
            self.assertEqual(
                {
                    "",
                    "0file",
                    "1top-dir",
                    "1top-dir/0file-in-1topdir",
                    "2utf\u1234file",
                },
                set(tree.all_versioned_paths()),
            )
            # note that the order of the paths and fileids is deliberately
            # mismatched to ensure that the result order is path based.
            self.assertEqual(
                [
                    ("", "directory"),
                    ("0file", "file"),
                    ("1top-dir", "directory"),
                    ("2utf\u1234file", "file"),
                    ("1top-dir/0file-in-1topdir", "file"),
                ],
                [(path, node.kind) for path, node in tree.iter_entries_by_dir()],
            )

    def _create_tree_with_utf8(self, tree):
        self.requireFeature(features.UnicodeFilenameFeature)

        # We avoid combining characters in file names here, normalization
        # checks (as performed by some file systems (OSX) are outside the scope
        # of these tests).  We use the euro sign \N{Euro Sign} or \u20ac in
        # unicode strings or '\xe2\x82\ac' (its utf-8 encoding) in raw strings.
        paths = [
            "",
            "fo\N{EURO SIGN}o",
            "ba\N{EURO SIGN}r/",
            "ba\N{EURO SIGN}r/ba\N{EURO SIGN}z",
        ]
        # bzr itself does not create unicode file ids, but we want them for
        # testing.
        file_ids = [
            b"TREE_ROOT",
            b"fo\xe2\x82\xaco-id",
            b"ba\xe2\x82\xacr-id",
            b"ba\xe2\x82\xacz-id",
        ]
        self.build_tree(paths[1:])
        if not tree.is_versioned(""):
            # Some trees do not have a root yet.
            tree.add(paths, ids=file_ids)
        else:
            # Some trees will already have a root
            if tree.supports_setting_file_ids():
                tree.set_root_id(file_ids[0])
                tree.add(paths[1:], ids=file_ids[1:])
            else:
                tree.add(paths[1:])
        if tree.branch.repository._format.supports_setting_revision_ids:
            try:
                tree.commit("in\xedtial", rev_id="r\xe9v-1".encode())
            except errors.NonAsciiRevisionId:
                raise TestSkipped("non-ascii revision ids not supported")
        else:
            tree.commit("in\xedtial")

        return tree

    def test_tree_with_utf8(self):
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            raise TestNotApplicable("format does not support custom file ids")
        self._create_tree_with_utf8(tree)

        tree = self.workingtree_to_test_tree(tree)

        revision_id = "r\xe9v-1".encode()
        root_id = b"TREE_ROOT"
        bar_id = "ba\N{EURO SIGN}r-id".encode()
        foo_id = "fo\N{EURO SIGN}o-id".encode()
        baz_id = "ba\N{EURO SIGN}z-id".encode()
        path_and_ids = [
            ("", root_id, None, None),
            ("ba\N{EURO SIGN}r", bar_id, root_id, revision_id),
            ("fo\N{EURO SIGN}o", foo_id, root_id, revision_id),
            ("ba\N{EURO SIGN}r/ba\N{EURO SIGN}z", baz_id, bar_id, revision_id),
        ]
        with tree.lock_read():
            path_entries = list(tree.iter_entries_by_dir())

        for expected, (path, ie) in zip(path_and_ids, path_entries):
            self.assertEqual(expected[0], path)  # Paths should match
            self.assertIsInstance(path, str)
            self.assertEqual(expected[1], ie.file_id)
            self.assertIsInstance(ie.file_id, bytes)
            self.assertEqual(expected[2], ie.parent_id)
            if expected[2] is not None:
                self.assertIsInstance(ie.parent_id, bytes)
            # WorkingTree's return None for the last modified revision
            if ie.revision is not None:
                self.assertIsInstance(ie.revision, bytes)
                if expected[0] != "":
                    # Some trees will preserve the revision id of the tree root,
                    # but not all will
                    self.assertEqual(revision_id, ie.revision)
        self.assertEqual(len(path_and_ids), len(path_entries))
        get_revision_id = getattr(tree, "get_revision_id", None)
        if get_revision_id is not None:
            self.assertIsInstance(get_revision_id(), bytes)
        last_revision = getattr(tree, "last_revision", None)
        if last_revision is not None:
            self.assertIsInstance(last_revision(), bytes)

    def test_tree_with_merged_utf8(self):
        wt = self.make_branch_and_tree(".")

        self._create_tree_with_utf8(wt)

        tree2 = wt.controldir.sprout("tree2").open_workingtree()
        self.build_tree(["tree2/ba\N{EURO SIGN}r/qu\N{EURO SIGN}x"])
        if wt.supports_setting_file_ids():
            tree2.add(
                ["ba\N{EURO SIGN}r/qu\N{EURO SIGN}x"],
                ids=["qu\N{EURO SIGN}x-id".encode()],
            )
        else:
            tree2.add(["ba\N{EURO SIGN}r/qu\N{EURO SIGN}x"])
        if wt.branch.repository._format.supports_setting_revision_ids:
            tree2.commit("to m\xe9rge", rev_id="r\xe9v-2".encode())
        else:
            tree2.commit("to m\xe9rge")

        self.assertTrue(tree2.is_versioned("ba\N{EURO SIGN}r/qu\N{EURO SIGN}x"))
        wt.merge_from_branch(tree2.branch)
        self.assertTrue(wt.is_versioned("ba\N{EURO SIGN}r/qu\N{EURO SIGN}x"))

        if wt.branch.repository._format.supports_setting_revision_ids:
            wt.commit("m\xe9rge", rev_id="r\xe9v-3".encode())
        else:
            wt.commit("m\xe9rge")
        tree = self.workingtree_to_test_tree(wt)

        revision_id_1 = "r\xe9v-1".encode()
        revision_id_2 = "r\xe9v-2".encode()
        root_id = b"TREE_ROOT"
        bar_id = "ba\N{EURO SIGN}r-id".encode()
        foo_id = "fo\N{EURO SIGN}o-id".encode()
        baz_id = "ba\N{EURO SIGN}z-id".encode()
        qux_id = "qu\N{EURO SIGN}x-id".encode()
        path_and_ids = [
            ("", root_id, None, None),
            ("ba\N{EURO SIGN}r", bar_id, root_id, revision_id_1),
            ("fo\N{EURO SIGN}o", foo_id, root_id, revision_id_1),
            ("ba\N{EURO SIGN}r/ba\N{EURO SIGN}z", baz_id, bar_id, revision_id_1),
            ("ba\N{EURO SIGN}r/qu\N{EURO SIGN}x", qux_id, bar_id, revision_id_2),
        ]
        with tree.lock_read():
            path_entries = list(tree.iter_entries_by_dir())

        for (epath, efid, eparent, _erev), (path, ie) in zip(
            path_and_ids, path_entries
        ):
            self.assertEqual(epath, path)  # Paths should match
            self.assertIsInstance(path, str)
            self.assertIsInstance(ie.file_id, bytes)
            if wt.supports_setting_file_ids():
                self.assertEqual(efid, ie.file_id)
                self.assertEqual(eparent, ie.parent_id)
            if eparent is not None:
                self.assertIsInstance(ie.parent_id, bytes)
        self.assertEqual(
            len(path_and_ids),
            len(path_entries),
            "{!r} vs {!r}".format(
                [p for (p, f, pf, r) in path_and_ids], [p for (p, e) in path_entries]
            ),
        )
        get_revision_id = getattr(tree, "get_revision_id", None)
        if get_revision_id is not None:
            self.assertIsInstance(get_revision_id(), bytes)
        last_revision = getattr(tree, "last_revision", None)
        if last_revision is not None:
            self.assertIsInstance(last_revision(), bytes)

    def skip_if_no_reference(self, tree):
        if not getattr(tree, "supports_tree_reference", lambda: False)():
            raise TestNotApplicable("Tree references not supported")

    def create_nested(self):
        work_tree = self.make_branch_and_tree("wt")
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree("wt/subtree")
            self.build_tree(["wt/subtree/a"])
            subtree.add(["a"])
            subtree.commit("foo")
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_iter_entries_with_unfollowed_reference(self):
        tree, subtree = self.create_nested()
        expected = [("", "directory"), ("subtree", "tree-reference")]
        with tree.lock_read():
            path_entries = list(tree.iter_entries_by_dir(recurse_nested=False))
            actual = [(path, ie.kind) for path, ie in path_entries]
        self.assertEqual(expected, actual)

    def test_iter_entries_with_followed_reference(self):
        tree, subtree = self.create_nested()
        expected = [("", "directory"), ("subtree", "directory"), ("subtree/a", "file")]
        with tree.lock_read():
            path_entries = list(tree.iter_entries_by_dir(recurse_nested=True))
            actual = [(path, ie.kind) for path, ie in path_entries]
        self.assertEqual(expected, actual)

    def test_iter_entries_with_missing_reference(self):
        tree, subtree = self.create_nested()
        shutil.rmtree("wt/subtree")
        with tree.lock_read():
            self.assertRaises(
                MissingNestedTree, list, tree.iter_entries_by_dir(recurse_nested=True)
            )



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the 'checkout' CLI command."""

import os

from breezy import branch as _mod_branch
from breezy import controldir, errors, workingtree
from breezy.bzr import bzrdir
from breezy.tests import TestCaseWithTransport
from breezy.tests.features import HardlinkFeature


class TestCheckout(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        tree = controldir.ControlDir.create_standalone_workingtree("branch")
        self.rev1 = tree.commit("1", allow_pointless=True)
        self.build_tree(["branch/added_in_2"])
        tree.add("added_in_2")
        self.rev2 = tree.commit("2")

    def test_checkout_makes_bound_branch(self):
        self.run_bzr("checkout branch checkout")
        # if we have a checkout, the branch base should be 'branch'
        source = controldir.ControlDir.open("branch")
        result = controldir.ControlDir.open("checkout")
        self.assertEqual(
            source.open_branch().controldir.root_transport.base,
            result.open_branch().get_bound_location(),
        )

    def test_checkout_light_makes_checkout(self):
        self.run_bzr("checkout --lightweight branch checkout")
        # if we have a checkout, the branch base should be 'branch'
        source = controldir.ControlDir.open("branch")
        result = controldir.ControlDir.open("checkout")
        self.assertEqual(
            source.open_branch().controldir.root_transport.base,
            result.open_branch().controldir.root_transport.base,
        )

    def test_checkout_dash_r(self):
        out, err = self.run_bzr(["checkout", "-r", "-2", "branch", "checkout"])
        # the working tree should now be at revision '1' with the content
        # from 1.
        result = controldir.ControlDir.open("checkout")
        self.assertEqual([self.rev1], result.open_workingtree().get_parent_ids())
        self.assertPathDoesNotExist("checkout/added_in_2")

    def test_checkout_light_dash_r(self):
        out, err = self.run_bzr(
            ["checkout", "--lightweight", "-r", "-2", "branch", "checkout"]
        )
        # the working tree should now be at revision '1' with the content
        # from 1.
        result = controldir.ControlDir.open("checkout")
        self.assertEqual([self.rev1], result.open_workingtree().get_parent_ids())
        self.assertPathDoesNotExist("checkout/added_in_2")

    def test_checkout_into_empty_dir(self):
        self.make_controldir("checkout")
        out, err = self.run_bzr(["checkout", "branch", "checkout"])
        result = controldir.ControlDir.open("checkout")
        result.open_workingtree()
        result.open_branch()

    def test_checkout_reconstitutes_working_trees(self):
        # doing a 'brz checkout' in the directory of a branch with no tree
        # or a 'brz checkout path' with path the name of a directory with
        # a branch with no tree will reconsistute the tree.
        os.mkdir("treeless-branch")
        branch = controldir.ControlDir.create_branch_convenience(
            "treeless-branch", force_new_tree=False, format=bzrdir.BzrDirMetaFormat1()
        )
        # check no tree was created
        self.assertRaises(errors.NoWorkingTree, branch.controldir.open_workingtree)
        out, err = self.run_bzr("checkout treeless-branch")
        # we should have a tree now
        branch.controldir.open_workingtree()
        # with no diff
        out, err = self.run_bzr("diff treeless-branch")

        # now test with no parameters
        branch = controldir.ControlDir.create_branch_convenience(
            ".", force_new_tree=False, format=bzrdir.BzrDirMetaFormat1()
        )
        # check no tree was created
        self.assertRaises(errors.NoWorkingTree, branch.controldir.open_workingtree)
        out, err = self.run_bzr("checkout")
        # we should have a tree now
        branch.controldir.open_workingtree()
        # with no diff
        out, err = self.run_bzr("diff")

    def _test_checkout_existing_dir(self, lightweight):
        source = self.make_branch_and_tree("source")
        self.build_tree_contents(
            [
                ("source/file1", b"content1"),
                ("source/file2", b"content2"),
            ]
        )
        source.add(["file1", "file2"])
        source.commit("added files")
        self.build_tree_contents(
            [
                ("target/", b""),
                ("target/file1", b"content1"),
                ("target/file2", b"content3"),
            ]
        )
        cmd = ["checkout", "source", "target"]
        if lightweight:
            cmd.append("--lightweight")
        self.run_bzr("checkout source target")
        # files with unique content should be moved
        self.assertPathExists("target/file2.moved")
        # files with content matching tree should not be moved
        self.assertPathDoesNotExist("target/file1.moved")

    def test_checkout_existing_dir_heavy(self):
        self._test_checkout_existing_dir(False)

    def test_checkout_existing_dir_lightweight(self):
        self._test_checkout_existing_dir(True)

    def test_checkout_in_branch_with_r(self):
        branch = _mod_branch.Branch.open("branch")
        branch.controldir.destroy_workingtree()
        self.run_bzr("checkout -r 1", working_dir="branch")
        tree = workingtree.WorkingTree.open("branch")
        self.assertEqual(self.rev1, tree.last_revision())
        branch.controldir.destroy_workingtree()
        self.run_bzr("checkout -r 0", working_dir="branch")
        self.assertEqual(b"null:", tree.last_revision())

    def test_checkout_files_from(self):
        _mod_branch.Branch.open("branch")
        self.run_bzr(["checkout", "branch", "branch2", "--files-from", "branch"])

    def test_checkout_hardlink(self):
        self.requireFeature(HardlinkFeature(self.test_dir))
        source = self.make_branch_and_tree("source")
        self.build_tree(["source/file1"])
        source.add("file1")
        source.commit("added file")
        out, err = self.run_bzr("checkout source target --hardlink")
        source_stat = os.stat("source/file1")
        target_stat = os.stat("target/file1")
        self.assertEqual(source_stat, target_stat)

    def test_checkout_hardlink_files_from(self):
        self.requireFeature(HardlinkFeature(self.test_dir))
        source = self.make_branch_and_tree("source")
        self.build_tree(["source/file1"])
        source.add("file1")
        source.commit("added file")
        source.controldir.sprout("second")
        out, err = self.run_bzr("checkout source target --hardlink --files-from second")
        second_stat = os.stat("second/file1")
        target_stat = os.stat("target/file1")
        self.assertEqual(second_stat, target_stat)

    def test_colo_checkout(self):
        source = self.make_branch_and_tree("source", format="development-colo")
        self.build_tree(["source/file1"])
        source.add("file1")
        source.commit("added file")
        target = source.controldir.sprout(
            "file:second,branch=somebranch", create_tree_if_local=False
        )
        out, err = self.run_bzr(
            "checkout file:,branch=somebranch .", working_dir="second"
        )
        # We should always be creating a lighweight checkout for colocated
        # branches.
        self.assertEqual(
            target.open_branch(name="somebranch").user_url,
            target.get_branch_reference(name=""),
        )



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys
import tempfile

from .. import controldir, errors, merge_directive, osutils
from ..bzr import generate_ids
from ..bzr.conflicts import ContentsConflict, PathConflict, TextConflict
from ..merge import Diff3Merger, Merge3Merger, Merger, WeaveMerger
from ..osutils import getcwd, pathjoin
from ..workingtree import WorkingTree
from . import TestCaseWithTransport, TestSkipped


class MergeBuilder:
    def __init__(self, dir=None):
        self.dir = tempfile.mkdtemp(prefix="merge-test", dir=dir)
        self.tree_root = generate_ids.gen_root_id()

        def wt(name):
            path = pathjoin(self.dir, name)
            os.mkdir(path)
            wt = controldir.ControlDir.create_standalone_workingtree(path)
            # the tests perform pulls, so need a branch that is writeable.
            wt.lock_write()
            if wt.supports_file_ids:
                wt.set_root_id(self.tree_root)
            wt.flush()
            tt = wt.transform()
            return wt, tt

        self.base, self.base_tt = wt("base")
        self.this, self.this_tt = wt("this")
        self.other, self.other_tt = wt("other")

    def root(self):
        ret = []
        for tt in [self.this_tt, self.base_tt, self.other_tt]:
            ret.append(tt.trans_id_tree_path(""))
        return tuple(ret)

    def get_cset_path(self, parent, name):
        if name is None:
            if parent is not None:
                raise AssertionError()
            return None
        return pathjoin(self.cset.entries[parent].path, name)

    def add_file(
        self,
        parent,
        name,
        contents,
        executable,
        this=True,
        base=True,
        other=True,
        file_id=None,
    ):
        ret = []
        for i, (option, tt) in enumerate(self.selected_transforms(this, base, other)):
            if option is True:
                trans_id = tt.new_file(
                    name, parent[i], [contents], executable=executable, file_id=file_id
                )
            else:
                trans_id = None
            ret.append(trans_id)
        return tuple(ret)

    def merge(self, merge_type=Merge3Merger, interesting_files=None, **kwargs):
        merger = self.make_merger(merge_type, interesting_files, **kwargs)
        merger.do_merge()
        return merger.cooked_conflicts

    def make_preview_transform(self):
        merger = self.make_merger(Merge3Merger, None, this_revision_tree=True)
        return merger.make_preview_transform()

    def make_merger(
        self, merge_type, interesting_files, this_revision_tree=False, **kwargs
    ):
        self.base_tt.apply()
        self.base.commit("base commit")
        for tt, wt in ((self.this_tt, self.this), (self.other_tt, self.other)):
            # why does this not do wt.pull() ?
            wt.branch.pull(self.base.branch)
            wt.set_parent_ids([wt.branch.last_revision()])
            wt.flush()
            # We maintain a write lock, so make sure changes are flushed to
            # disk first
            tt.apply()
            wt.commit("branch commit")
            wt.flush()
            if wt.branch.last_revision_info()[0] != 2:
                raise AssertionError()
        self.this.branch.fetch(self.other.branch)
        other_basis = self.other.branch.basis_tree()
        if this_revision_tree:
            self.this.commit("message")
            this_tree = self.this.basis_tree()
        else:
            this_tree = self.this
        merger = merge_type(
            this_tree,
            self.this,
            self.base,
            other_basis,
            interesting_files=interesting_files,
            do_merge=False,
            this_branch=self.this.branch,
            **kwargs,
        )
        return merger

    def list_transforms(self):
        return [self.this_tt, self.base_tt, self.other_tt]

    def selected_transforms(self, this, base, other):
        pairs = [(this, self.this_tt), (base, self.base_tt), (other, self.other_tt)]
        return [(v, tt) for (v, tt) in pairs if v is not None]

    def add_symlink(self, parent, name, contents, file_id=None):
        ret = []
        for i, tt in enumerate(self.list_transforms()):
            trans_id = tt.new_symlink(name, parent[i], contents, file_id=file_id)
            ret.append(trans_id)
        return ret

    def remove_file(self, trans_ids, base=False, this=False, other=False):
        for trans_id, (option, tt) in zip(
            trans_ids, self.selected_transforms(this, base, other)
        ):
            if option is True:
                tt.cancel_creation(trans_id)
                tt.cancel_versioning(trans_id)
                tt.set_executability(None, trans_id)

    def add_dir(self, parent, name, this=True, base=True, other=True, file_id=None):
        ret = []
        for i, (option, tt) in enumerate(self.selected_transforms(this, base, other)):
            if option is True:
                trans_id = tt.new_directory(name, parent[i], file_id)
            else:
                trans_id = None
            ret.append(trans_id)
        return tuple(ret)

    def change_name(self, trans_ids, base=None, this=None, other=None):
        for val, tt, trans_id in (
            (base, self.base_tt, trans_ids[0]),
            (this, self.this_tt, trans_ids[1]),
            (other, self.other_tt, trans_ids[2]),
        ):
            if val is None:
                continue
            parent_id = tt.final_parent(trans_id)
            tt.adjust_path(val, parent_id, trans_id)

    def change_parent(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (parent, tt) in zip(
            trans_ids, self.selected_transforms(this, base, other)
        ):
            parent_id = tt.trans_id_file_id(parent)
            tt.adjust_path(tt.final_name(trans_id), parent_id, trans_id)

    def change_contents(self, trans_id, base=None, this=None, other=None):
        for trans_id, (contents, tt) in zip(
            trans_id, self.selected_transforms(this, base, other)
        ):
            tt.cancel_creation(trans_id)
            tt.create_file([contents], trans_id)

    def change_target(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (target, tt) in zip(
            trans_ids, self.selected_transforms(this, base, other)
        ):
            tt.cancel_creation(trans_id)
            tt.create_symlink(target, trans_id)

    def change_perms(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (executability, tt) in zip(
            trans_ids, self.selected_transforms(this, base, other)
        ):
            tt.set_executability(None, trans_id)
            tt.set_executability(executability, trans_id)

    def apply_inv_change(self, inventory_change, orig_inventory):
        orig_inventory_by_path = {}
        for file_id, path in orig_inventory.items():
            orig_inventory_by_path[path] = file_id

        def parent_id(file_id):
            try:
                parent_dir = os.path.dirname(orig_inventory[file_id])
            except:
                print(file_id)
                raise
            if parent_dir == "":
                return None
            return orig_inventory_by_path[parent_dir]

        def new_path(file_id):
            if fild_id in inventory_change:
                return inventory_change[file_id]
            else:
                parent = parent_id(file_id)
                if parent is None:
                    return orig_inventory[file_id]
                dirname = new_path(parent)
                return pathjoin(dirname, os.path.basename(orig_inventory[file_id]))

        new_inventory = {}
        for file_id in orig_inventory:
            path = new_path(file_id)
            if path is None:
                continue
            new_inventory[file_id] = path

        for file_id, path in inventory_change.items():
            if file_id in orig_inventory:
                continue
            new_inventory[file_id] = path
        return new_inventory

    def unlock(self):
        self.base.unlock()
        self.this.unlock()
        self.other.unlock()

    def cleanup(self):
        self.unlock()
        osutils.rmtree(self.dir)


class MergeTest(TestCaseWithTransport):
    def test_change_name(self):
        """Test renames."""
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"hello1", True, file_id=b"1")
        builder.change_name(name1, other="name2")
        name3 = builder.add_file(builder.root(), "name3", b"hello2", True, file_id=b"2")
        builder.change_name(name3, base="name4")
        name5 = builder.add_file(builder.root(), "name5", b"hello3", True, file_id=b"3")
        builder.change_name(name5, this="name6")
        builder.merge()
        builder.cleanup()
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(
            builder.root(), "name1", b"hello1", False, file_id=b"1"
        )
        builder.change_name(name1, other="name2", this="name3")
        conflicts = builder.merge()
        self.assertEqual(conflicts, [PathConflict("name3", "name2", b"1")])
        builder.cleanup()

    def test_merge_one(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"hello1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4")
        name2 = builder.add_file(builder.root(), "name2", b"hello1", True, file_id=b"2")
        builder.change_contents(name2, other=b"text4")
        builder.merge(interesting_files=["name1"])
        self.assertEqual(builder.this.get_file("name1").read(), b"text4")
        self.assertEqual(builder.this.get_file("name2").read(), b"hello1")
        builder.cleanup()

    def test_file_moves(self):
        """Test moves."""
        builder = MergeBuilder(getcwd())
        dir1 = builder.add_dir(builder.root(), "dir1", file_id=b"1")
        builder.add_dir(builder.root(), "dir2", file_id=b"2")
        file1 = builder.add_file(dir1, "file1", b"hello1", True, file_id=b"3")
        file2 = builder.add_file(dir1, "file2", b"hello2", True, file_id=b"4")
        file3 = builder.add_file(dir1, "file3", b"hello3", True, file_id=b"5")
        builder.change_parent(file1, other=b"2")
        builder.change_parent(file2, this=b"2")
        builder.change_parent(file3, base=b"2")
        builder.merge()
        builder.cleanup()

        builder = MergeBuilder(getcwd())
        dir1 = builder.add_dir(builder.root(), "dir1", file_id=b"1")
        builder.add_dir(builder.root(), "dir2", file_id=b"2")
        builder.add_dir(builder.root(), "dir3", file_id=b"3")
        file1 = builder.add_file(dir1, "file1", b"hello1", False, file_id=b"4")
        builder.change_parent(file1, other=b"2", this=b"3")
        conflicts = builder.merge()
        path2 = pathjoin("dir2", "file1")
        path3 = pathjoin("dir3", "file1")
        self.assertEqual(conflicts, [PathConflict(path3, path2, b"4")])
        builder.cleanup()

    def test_contents_merge(self):
        """Test merge3 merging."""
        self.do_contents_test(Merge3Merger)

    def test_contents_merge2(self):
        """Test diff3 merging."""
        if sys.platform == "win32":
            raise TestSkipped(
                "diff3 does not have --binary flag and therefore always fails on win32"
            )
        try:
            self.do_contents_test(Diff3Merger)
        except errors.NoDiff3:
            raise TestSkipped("diff3 not available")

    def test_contents_merge3(self):
        """Test diff3 merging."""
        self.do_contents_test(WeaveMerger)

    def test_reprocess_weave(self):
        # Reprocess works on weaves, and behaves as expected
        builder = MergeBuilder(getcwd())
        blah = builder.add_file(builder.root(), "blah", b"a", False, file_id=b"a")
        builder.change_contents(blah, this=b"b\nc\nd\ne\n", other=b"z\nc\nd\ny\n")
        builder.merge(WeaveMerger, reprocess=True)
        expected = b"""<<<<<<< TREE
b
=======
z
>>>>>>> MERGE-SOURCE
c
d
<<<<<<< TREE
e
=======
y
>>>>>>> MERGE-SOURCE
"""
        self.assertEqualDiff(builder.this.get_file_text("blah"), expected)
        builder.cleanup()

    def do_contents_test(self, merge_factory):
        """Test merging with specified ContentsChange factory."""
        builder = self.contents_test_success(merge_factory)
        builder.cleanup()
        self.contents_test_conflicts(merge_factory)

    def contents_test_success(self, merge_factory):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4")
        name3 = builder.add_file(builder.root(), "name3", b"text2", False, file_id=b"2")
        builder.change_contents(name3, base=b"text5")
        builder.add_file(builder.root(), "name5", b"text3", True, file_id=b"3")
        name6 = builder.add_file(builder.root(), "name6", b"text4", True, file_id=b"4")
        builder.remove_file(name6, base=True)
        name7 = builder.add_file(
            builder.root(), "name7", b"a\nb\nc\nd\ne\nf\n", True, file_id=b"5"
        )
        builder.change_contents(
            name7, other=b"a\nz\nc\nd\ne\nf\n", this=b"a\nb\nc\nd\ne\nz\n"
        )
        conflicts = builder.merge(merge_factory)
        try:
            self.assertEqual([], conflicts)
            self.assertEqual(b"text4", builder.this.get_file("name1").read())
            self.assertEqual(b"text2", builder.this.get_file("name3").read())
            self.assertEqual(
                b"a\nz\nc\nd\ne\nz\n", builder.this.get_file("name7").read()
            )
            self.assertTrue(builder.this.is_executable("name1"))
            self.assertFalse(builder.this.is_executable("name3"))
            self.assertTrue(builder.this.is_executable("name5"))
        except:
            builder.unlock()
            raise
        return builder

    def contents_test_conflicts(self, merge_factory):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4", this=b"text3")
        name2 = builder.add_file(builder.root(), "name2", b"text1", True, file_id=b"2")
        builder.change_contents(name2, other=b"\x00", this=b"text3")
        name3 = builder.add_file(builder.root(), "name3", b"text5", False, file_id=b"3")
        builder.change_perms(name3, this=True)
        builder.change_contents(name3, this=b"moretext")
        builder.remove_file(name3, other=True)
        conflicts = builder.merge(merge_factory)
        self.assertEqual(
            conflicts,
            [
                TextConflict("name1", file_id=b"1"),
                ContentsConflict("name2", file_id=b"2"),
                ContentsConflict("name3", file_id=b"3"),
            ],
        )
        with builder.this.get_file(builder.this.id2path(b"2")) as f:
            self.assertEqual(f.read(), b"\x00")
        builder.cleanup()

    def test_symlink_conflicts(self):
        if sys.platform != "win32":
            builder = MergeBuilder(getcwd())
            name2 = builder.add_symlink(
                builder.root(), "name2", "target1", file_id=b"2"
            )
            builder.change_target(name2, other="target4", base="text3")
            conflicts = builder.merge()
            self.assertEqual(conflicts, [ContentsConflict("name2", file_id=b"2")])
            builder.cleanup()

    def test_symlink_merge(self):
        if sys.platform != "win32":
            builder = MergeBuilder(getcwd())
            name1 = builder.add_symlink(
                builder.root(), "name1", "target1", file_id=b"1"
            )
            name2 = builder.add_symlink(
                builder.root(), "name2", "target1", file_id=b"2"
            )
            name3 = builder.add_symlink(
                builder.root(), "name3", "target1", file_id=b"3"
            )
            builder.change_target(name1, this=b"target2")
            builder.change_target(name2, base=b"target2")
            builder.change_target(name3, other=b"target2")
            builder.merge()
            self.assertEqual(builder.this.get_symlink_target("name1"), "target2")
            self.assertEqual(builder.this.get_symlink_target("name2"), "target1")
            self.assertEqual(builder.this.get_symlink_target("name3"), "target2")
            builder.cleanup()

    def test_no_passive_add(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.remove_file(name1, this=True)
        builder.merge()
        builder.cleanup()

    def test_perms_merge(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_perms(name1, other=False)
        name2 = builder.add_file(builder.root(), "name2", b"text2", True, file_id=b"2")
        builder.change_perms(name2, base=False)
        name3 = builder.add_file(builder.root(), "name3", b"text3", True, file_id=b"3")
        builder.change_perms(name3, this=False)
        name4 = builder.add_file(builder.root(), "name4", b"text4", False, file_id=b"4")
        builder.change_perms(name4, this=True)
        builder.remove_file(name4, base=True)
        builder.merge()
        self.assertIs(builder.this.is_executable("name1"), False)
        self.assertIs(builder.this.is_executable("name2"), True)
        self.assertIs(builder.this.is_executable("name3"), False)
        builder.cleanup()

    def test_new_suffix(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text3")
        builder.add_file(builder.root(), "name1.new", b"text2", True, file_id=b"2")
        builder.merge()
        os.lstat(builder.this.abspath("name1.new"))
        builder.cleanup()

    def test_spurious_conflict(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", False, file_id=b"1")
        builder.remove_file(name1, other=True)
        builder.add_file(
            builder.root(),
            "name1",
            b"text1",
            False,
            this=False,
            base=False,
            file_id=b"2",
        )
        conflicts = builder.merge()
        self.assertEqual(conflicts, [])
        builder.cleanup()

    def test_merge_one_renamed(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(
            builder.root(), "name1", b"text1a", False, file_id=b"1"
        )
        builder.change_name(name1, this="name2")
        builder.change_contents(name1, other=b"text2")
        builder.merge(interesting_files=["name2"])
        self.assertEqual(b"text2", builder.this.get_file("name2").read())
        builder.cleanup()


class FunctionalMergeTest(TestCaseWithTransport):
    def test_trivial_star_merge(self):
        """Test that merges in a star shape Just Work."""
        # John starts a branch
        self.build_tree(("original/", "original/file1", "original/file2"))
        tree = self.make_branch_and_tree("original")
        branch = tree.branch
        tree.smart_add(["original"])
        tree.commit("start branch.", verbose=False)
        # Mary branches it.
        self.build_tree(("mary/",))
        branch.controldir.clone("mary")
        # Now John commits a change
        with open("original/file1", "w") as f:
            f.write("John\n")
        tree.commit("change file1")
        # Mary does too
        mary_tree = WorkingTree.open("mary")
        with open("mary/file2", "w") as f:
            f.write("Mary\n")
        mary_tree.commit("change file2")
        # john should be able to merge with no conflicts.
        tree.merge_from_branch(mary_tree.branch)
        with open("original/file1") as f:
            self.assertEqual("John\n", f.read())
        with open("original/file2") as f:
            self.assertEqual("Mary\n", f.read())

    def test_conflicts(self):
        wta = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/file", b"contents\n")])
        wta.add("file")
        wta.commit("base revision", allow_pointless=False)
        d_b = wta.branch.controldir.clone("b")
        self.build_tree_contents([("a/file", b"other contents\n")])
        wta.commit("other revision", allow_pointless=False)
        self.build_tree_contents([("b/file", b"this contents contents\n")])
        wtb = d_b.open_workingtree()
        wtb.commit("this revision", allow_pointless=False)
        self.assertEqual(1, len(wtb.merge_from_branch(wta.branch)))
        self.assertPathExists("b/file.THIS")
        self.assertPathExists("b/file.BASE")
        self.assertPathExists("b/file.OTHER")
        wtb.revert()
        self.assertEqual(
            1, len(wtb.merge_from_branch(wta.branch, merge_type=WeaveMerger))
        )
        self.assertPathExists("b/file")
        self.assertPathExists("b/file.THIS")
        self.assertPathExists("b/file.BASE")
        self.assertPathExists("b/file.OTHER")

    def test_weave_conflicts_not_in_base(self):
        builder = self.make_branch_builder("source")
        builder.start_series()
        # See bug #494197
        #  A        base revision (before criss-cross)
        #  |\
        #  B C      B does nothing, C adds 'foo'
        #  |X|
        #  D E      D and E modify foo in incompatible ways
        #
        # Merging will conflict, with C as a clean base text. However, the
        # current code uses A as the global base and 'foo' doesn't exist there.
        # It isn't trivial to create foo.BASE because it tries to look up
        # attributes like 'executable' in A.
        a_id = builder.build_snapshot(
            None, [("add", ("", b"TREE_ROOT", "directory", None))]
        )
        b_id = builder.build_snapshot([a_id], [])
        c_id = builder.build_snapshot(
            [a_id], [("add", ("foo", b"foo-id", "file", b"orig\ncontents\n"))]
        )
        d_id = builder.build_snapshot(
            [b_id, c_id],
            [("add", ("foo", b"foo-id", "file", b"orig\ncontents\nand D\n"))],
        )
        builder.build_snapshot(
            [c_id, b_id], [("modify", ("foo", b"orig\ncontents\nand E\n"))]
        )
        builder.finish_series()
        tree = builder.get_branch().create_checkout("tree", lightweight=True)
        self.assertEqual(
            1,
            len(
                tree.merge_from_branch(
                    tree.branch, to_revision=d_id, merge_type=WeaveMerger
                )
            ),
        )
        self.assertPathExists("tree/foo.THIS")
        self.assertPathExists("tree/foo.OTHER")
        self.expectFailure(
            "fail to create .BASE in some criss-cross merges",
            self.assertPathExists,
            "tree/foo.BASE",
        )
        self.assertPathExists("tree/foo.BASE")

    def test_merge_unrelated(self):
        """Sucessfully merges unrelated branches with no common names."""
        wta = self.make_branch_and_tree("a")
        with open("a/a_file", "wb") as f:
            f.write(b"contents\n")
        wta.add("a_file")
        wta.commit("a_revision", allow_pointless=False)
        wtb = self.make_branch_and_tree("b")
        with open("b/b_file", "wb") as f:
            f.write(b"contents\n")
        wtb.add("b_file")
        b_rev = wtb.commit("b_revision", allow_pointless=False)
        wta.merge_from_branch(wtb.branch, b_rev, b"null:")
        self.assertTrue(os.path.lexists("a/b_file"))
        self.assertEqual([b_rev], wta.get_parent_ids()[1:])

    def test_merge_unrelated_conflicting(self):
        """Sucessfully merges unrelated branches with common names."""
        wta = self.make_branch_and_tree("a")
        with open("a/file", "wb") as f:
            f.write(b"contents\n")
        wta.add("file")
        wta.commit("a_revision", allow_pointless=False)
        wtb = self.make_branch_and_tree("b")
        with open("b/file", "wb") as f:
            f.write(b"contents\n")
        wtb.add("file")
        b_rev = wtb.commit("b_revision", allow_pointless=False)
        wta.merge_from_branch(wtb.branch, b_rev, b"null:")
        self.assertTrue(os.path.lexists("a/file"))
        self.assertTrue(os.path.lexists("a/file.moved"))
        self.assertEqual([b_rev], wta.get_parent_ids()[1:])

    def test_merge_deleted_conflicts(self):
        wta = self.make_branch_and_tree("a")
        with open("a/file", "wb") as f:
            f.write(b"contents\n")
        wta.add("file")
        wta.commit("a_revision", allow_pointless=False)
        self.run_bzr("branch a b")
        os.remove("a/file")
        wta.commit("removed file", allow_pointless=False)
        with open("b/file", "wb") as f:
            f.write(b"changed contents\n")
        wtb = WorkingTree.open("b")
        wtb.commit("changed file", allow_pointless=False)
        wtb.merge_from_branch(
            wta.branch, wta.branch.last_revision(), wta.branch.get_rev_id(1)
        )
        self.assertFalse(os.path.lexists("b/file"))

    def test_merge_metadata_vs_deletion(self):
        """Conflict deletion vs metadata change."""
        a_wt = self.make_branch_and_tree("a")
        with open("a/file", "wb") as f:
            f.write(b"contents\n")
        a_wt.add("file")
        a_wt.commit("r0")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        os.chmod("b/file", 0o755)
        os.remove("a/file")
        a_wt.commit("removed a")
        self.assertEqual(a_wt.branch.revno(), 2)
        self.assertFalse(os.path.exists("a/file"))
        b_wt.commit("exec a")
        a_wt.merge_from_branch(b_wt.branch, b_wt.last_revision(), b"null:")
        self.assertTrue(os.path.exists("a/file"))

    def test_merge_swapping_renames(self):
        a_wt = self.make_branch_and_tree("a")
        with open("a/un", "wb") as f:
            f.write(b"UN")
        with open("a/deux", "wb") as f:
            f.write(b"DEUX")
        a_wt.add("un")
        a_wt.add("deux")
        a_wt.commit("r0", rev_id=b"r0")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        b_wt.rename_one("un", "tmp")
        b_wt.rename_one("deux", "un")
        b_wt.rename_one("tmp", "deux")
        b_wt.commit("r1", rev_id=b"r1")
        self.assertEqual(
            0,
            len(
                a_wt.merge_from_branch(
                    b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
                )
            ),
        )
        self.assertPathExists("a/un")
        self.assertTrue("a/deux")
        self.assertFalse(os.path.exists("a/tmp"))
        with open("a/un") as f:
            self.assertEqual(f.read(), "DEUX")
        with open("a/deux") as f:
            self.assertEqual(f.read(), "UN")

    def test_merge_delete_and_add_same(self):
        a_wt = self.make_branch_and_tree("a")
        with open("a/file", "wb") as f:
            f.write(b"THIS")
        a_wt.add("file")
        a_wt.commit("r0")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        os.remove("b/file")
        b_wt.commit("r1")
        with open("b/file", "wb") as f:
            f.write(b"THAT")
        b_wt.add("file")
        b_wt.commit("r2")
        a_wt.merge_from_branch(
            b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
        )
        self.assertTrue(os.path.exists("a/file"))
        with open("a/file") as f:
            self.assertEqual(f.read(), "THAT")

    def test_merge_rename_before_create(self):
        """Rename before create.

        This case requires that you must not do creates
        before move-into-place:

        $ touch foo
        $ bzr add foo
        $ bzr commit
        $ bzr mv foo bar
        $ touch foo
        $ bzr add foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree("a")
        with open("a/foo", "wb") as f:
            f.write(b"A/FOO")
        a_wt.add("foo")
        a_wt.commit("added foo")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        b_wt.rename_one("foo", "bar")
        with open("b/foo", "wb") as f:
            f.write(b"B/FOO")
        b_wt.add("foo")
        b_wt.commit("moved foo to bar, added new foo")
        a_wt.merge_from_branch(
            b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
        )

    def test_merge_create_before_rename(self):
        """Create before rename, target parents before children.

        This case requires that you must not do move-into-place
        before creates, and that you must not do children after
        parents:

        $ touch foo
        $ bzr add foo
        $ bzr commit
        $ bzr mkdir bar
        $ bzr add bar
        $ bzr mv foo bar/foo
        $ bzr commit
        """
        os.mkdir("a")
        a_wt = self.make_branch_and_tree("a")
        with open("a/foo", "wb") as f:
            f.write(b"A/FOO")
        a_wt.add("foo")
        a_wt.commit("added foo")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        os.mkdir("b/bar")
        b_wt.add("bar")
        b_wt.rename_one("foo", "bar/foo")
        b_wt.commit("created bar dir, moved foo into bar")
        a_wt.merge_from_branch(
            b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
        )

    def test_merge_rename_to_temp_before_delete(self):
        """Rename to temp before delete, source children before parents.

        This case requires that you must not do deletes before
        move-out-of-the-way, and that you must not do children
        after parents:

        $ mkdir foo
        $ touch foo/bar
        $ bzr add foo/bar
        $ bzr commit
        $ bzr mv foo/bar bar
        $ rmdir foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree("a")
        os.mkdir("a/foo")
        with open("a/foo/bar", "wb") as f:
            f.write(b"A/FOO/BAR")
        a_wt.add("foo")
        a_wt.add("foo/bar")
        a_wt.commit("added foo/bar")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        b_wt.rename_one("foo/bar", "bar")
        os.rmdir("b/foo")
        b_wt.remove("foo")
        b_wt.commit("moved foo/bar to bar, deleted foo")
        a_wt.merge_from_branch(
            b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
        )

    def test_merge_delete_before_rename_to_temp(self):
        """Delete before rename to temp.

        This case requires that you must not do
        move-out-of-the-way before deletes:

        $ touch foo
        $ touch bar
        $ bzr add foo bar
        $ bzr commit
        $ rm foo
        $ bzr rm foo
        $ bzr mv bar foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree("a")
        with open("a/foo", "wb") as f:
            f.write(b"A/FOO")
        with open("a/bar", "wb") as f:
            f.write(b"A/BAR")
        a_wt.add("foo")
        a_wt.add("bar")
        a_wt.commit("added foo and bar")
        self.run_bzr("branch a b")
        b_wt = WorkingTree.open("b")
        os.unlink("b/foo")
        b_wt.remove("foo")
        b_wt.rename_one("bar", "foo")
        b_wt.commit("deleted foo, renamed bar to foo")
        a_wt.merge_from_branch(
            b_wt.branch, b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)
        )


class TestMerger(TestCaseWithTransport):
    def set_up_trees(self):
        this = self.make_branch_and_tree("this")
        this.commit("rev1", rev_id=b"rev1")
        other = this.controldir.sprout("other").open_workingtree()
        this.commit("rev2a", rev_id=b"rev2a")
        other.commit("rev2b", rev_id=b"rev2b")
        return this, other

    def test_from_revision_ids(self):
        this, other = self.set_up_trees()
        self.assertRaises(
            errors.NoSuchRevision, Merger.from_revision_ids, this, b"rev2b"
        )
        this.lock_write()
        self.addCleanup(this.unlock)
        merger = Merger.from_revision_ids(this, b"rev2b", other_branch=other.branch)
        self.assertEqual(b"rev2b", merger.other_rev_id)
        self.assertEqual(b"rev1", merger.base_rev_id)
        merger = Merger.from_revision_ids(
            this, b"rev2b", b"rev2a", other_branch=other.branch
        )
        self.assertEqual(b"rev2a", merger.base_rev_id)

    def test_from_uncommitted(self):
        this, other = self.set_up_trees()
        merger = Merger.from_uncommitted(this, other, None)
        self.assertIs(other, merger.other_tree)
        self.assertIs(None, merger.other_rev_id)
        self.assertEqual(b"rev2b", merger.base_rev_id)

    def prepare_for_merging(self):
        this, other = self.set_up_trees()
        other.commit("rev3", rev_id=b"rev3")
        this.lock_write()
        self.addCleanup(this.unlock)
        return this, other

    def test_from_mergeable(self):
        this, other = self.prepare_for_merging()
        md = merge_directive.MergeDirective2.from_objects(
            other.branch.repository, b"rev3", 0, 0, "this"
        )
        other.lock_read()
        self.addCleanup(other.unlock)
        merger, verified = Merger.from_mergeable(this, md)
        md.patch = None
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual("inapplicable", verified)
        self.assertEqual(b"rev3", merger.other_rev_id)
        self.assertEqual(b"rev1", merger.base_rev_id)
        md.base_revision_id = b"rev2b"
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual(b"rev2b", merger.base_rev_id)

    def test_from_mergeable_old_merge_directive(self):
        this, other = self.prepare_for_merging()
        other.lock_write()
        self.addCleanup(other.unlock)
        md = merge_directive.MergeDirective.from_objects(
            other.branch.repository, b"rev3", 0, 0, "this"
        )
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual(b"rev3", merger.other_rev_id)
        self.assertEqual(b"rev1", merger.base_rev_id)



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz whoami."""

from breezy import branch, config, errors, tests

from ..test_bedding import override_whoami


class TestWhoami(tests.TestCaseWithTransport):
    def assertWhoAmI(self, expected, *cmd_args, **kwargs):
        out, err = self.run_bzr(("whoami",) + cmd_args, **kwargs)
        self.assertEqual("", err)
        lines = out.splitlines()
        self.assertLength(1, lines)
        if isinstance(expected, bytes):
            expected = expected.decode(kwargs.get("encoding", "ascii"))
        self.assertEqual(expected, lines[0].rstrip())

    def test_whoami_no_args_no_conf(self):
        # this should always identify something, if only "john@localhost"
        out = self.run_bzr("whoami")[0]
        self.assertTrue(len(out) > 0)
        self.assertEqual(1, out.count("@"))

    def test_whoami_email_no_args(self):
        out = self.run_bzr("whoami --email")[0]
        self.assertTrue(len(out) > 0)
        self.assertEqual(1, out.count("@"))

    def test_whoami_email_arg(self):
        # whoami --email is mutually exclusive with any arguments
        out = self.run_bzr("whoami --email 'foo <foo@example.com>'", 3)[0]
        self.assertEqual("", out)

    def set_branch_email(self, b, email):
        b.get_config_stack().set("email", email)

    def test_whoami_branch(self):
        """Branch specific user identity works."""
        self.make_branch_and_tree(".")
        b = branch.Branch.open(".")
        self.set_branch_email(b, "Branch Identity <branch@identi.ty>")
        self.assertWhoAmI("Branch Identity <branch@identi.ty>")
        self.assertWhoAmI("branch@identi.ty", "--email")

        # Verify that the environment variable overrides the value
        # in the file
        self.overrideEnv("BRZ_EMAIL", "Different ID <other@environ.ment>")
        self.assertWhoAmI("Different ID <other@environ.ment>")
        self.assertWhoAmI("other@environ.ment", "--email")

    def test_whoami_utf8(self):
        """Verify that an identity can be in utf-8."""
        self.run_bzr(
            ["whoami", "Branch Identity \u20ac <branch@identi.ty>"], encoding="utf-8"
        )
        self.assertWhoAmI(
            b"Branch Identity \xe2\x82\xac <branch@identi.ty>", encoding="utf-8"
        )
        self.assertWhoAmI("branch@identi.ty", "--email")

    def test_whoami_ascii(self):
        """Verify that whoami doesn't totally break when in utf-8, using an ascii
        encoding.
        """
        self.make_branch_and_tree(".")
        b = branch.Branch.open(".")
        self.set_branch_email(b, "Branch Identity \u20ac <branch@identi.ty>")
        self.assertWhoAmI("Branch Identity ? <branch@identi.ty>", encoding="ascii")
        self.assertWhoAmI("branch@identi.ty", "--email", encoding="ascii")

    def test_warning(self):
        """Verify that a warning is displayed if no email is given."""
        self.make_branch_and_tree(".")
        display = self.run_bzr(["whoami", "Branch Identity"])[1]
        self.assertEqual(
            '"Branch Identity" does not seem to contain an '
            "email address.  This is allowed, but not "
            "recommended.\n",
            display,
        )

    def test_whoami_not_set(self):
        """Ensure whoami error if username is not set and not inferred."""
        override_whoami(self)
        out, err = self.run_bzr(["whoami"], 3)
        self.assertContainsRe(err, "Unable to determine your name")

    def test_whoami_directory(self):
        """Test --directory option."""
        wt = self.make_branch_and_tree("subdir")
        self.set_branch_email(wt.branch, "Branch Identity <branch@identi.ty>")
        self.assertWhoAmI("Branch Identity <branch@identi.ty>", "--directory", "subdir")
        self.run_bzr(
            [
                "whoami",
                "--directory",
                "subdir",
                "--branch",
                "Changed Identity <changed@identi.ty>",
            ]
        )
        # Refresh wt as 'whoami' modified it
        wt = wt.controldir.open_workingtree()
        c = wt.branch.get_config_stack()
        self.assertEqual("Changed Identity <changed@identi.ty>", c.get("email"))

    def test_whoami_remote_directory(self):
        """Test --directory option with a remote directory."""
        wt = self.make_branch_and_tree("subdir")
        self.set_branch_email(wt.branch, "Branch Identity <branch@identi.ty>")
        url = self.get_readonly_url() + "/subdir"
        self.assertWhoAmI("Branch Identity <branch@identi.ty>", "--directory", url)
        url = self.get_url("subdir")
        self.run_bzr(
            [
                "whoami",
                "--directory",
                url,
                "--branch",
                "Changed Identity <changed@identi.ty>",
            ]
        )
        # The identity has been set in the branch config (but not the global
        # config)
        c = branch.Branch.open(url).get_config_stack()
        self.assertEqual("Changed Identity <changed@identi.ty>", c.get("email"))
        # Ensuring that the value does not come from the breezy.conf file
        # itself requires some isolation setup
        override_whoami(self)
        global_conf = config.GlobalStack()
        self.assertRaises(errors.NoWhoami, global_conf.get, "email")

    def test_whoami_nonbranch_directory(self):
        """Test --directory mentioning a non-branch directory."""
        self.build_tree(["subdir/"])
        out, err = self.run_bzr("whoami --directory subdir", retcode=3)
        self.assertContainsRe(err, "ERROR: Not a branch")



# Copyright (C) 2006 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the GitMemoryTree class."""

from ... import errors
from ...transport import NoSuchFile
from . import TestCaseWithTransport


class TestMemoryTree(TestCaseWithTransport):
    def make_branch(self, path, format="git"):
        return super().make_branch(path, format=format)

    def make_branch_and_tree(self, path, format="git"):
        return super().make_branch_and_tree(path, format=format)

    def test_create_on_branch(self):
        """Creating a mutable tree on a trivial branch works."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        self.assertEqual(branch.controldir, tree.controldir)
        self.assertEqual(branch, tree.branch)
        self.assertEqual([], tree.get_parent_ids())

    def test_create_on_branch_with_content(self):
        """Creating a mutable tree on a non-trivial branch works."""
        wt = self.make_branch_and_tree("sometree")
        self.build_tree(["sometree/foo"])
        wt.add(["foo"])
        rev_id = wt.commit("first post")
        tree = wt.branch.create_memorytree()
        with tree.lock_read():
            self.assertEqual([rev_id], tree.get_parent_ids())
            self.assertEqual(b"contents of sometree/foo\n", tree.get_file("foo").read())

    def test_lock_tree_write(self):
        """Check we can lock_tree_write and unlock MemoryTrees."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        tree.lock_tree_write()
        tree.unlock()

    def test_lock_tree_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        tree.lock_read()
        self.assertRaises(errors.ReadOnlyError, tree.lock_tree_write)
        tree.unlock()

    def test_lock_write(self):
        """Check we can lock_write and unlock MemoryTrees."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        tree.lock_write()
        tree.unlock()

    def test_lock_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        tree.lock_read()
        self.assertRaises(errors.ReadOnlyError, tree.lock_write)
        tree.unlock()

    def test_add_with_kind(self):
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        tree.lock_write()
        tree.add(["", "afile", "adir"], ["directory", "file", "directory"])
        self.assertTrue(tree.is_versioned("afile"))
        self.assertFalse(tree.is_versioned("adir"))
        self.assertFalse(tree.has_filename("afile"))
        self.assertFalse(tree.has_filename("adir"))
        tree.unlock()

    def test_put_new_file(self):
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(["", "foo"], kinds=["directory", "file"])
            tree.put_file_bytes_non_atomic("foo", b"barshoom")
            self.assertEqual(b"barshoom", tree.get_file("foo").read())

    def test_put_existing_file(self):
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(["", "foo"], kinds=["directory", "file"])
            tree.put_file_bytes_non_atomic("foo", b"first-content")
            tree.put_file_bytes_non_atomic("foo", b"barshoom")
            self.assertEqual(b"barshoom", tree.get_file("foo").read())

    def test_add_in_subdir(self):
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add([""], ["directory"])
            tree.mkdir("adir")
            tree.put_file_bytes_non_atomic("adir/afile", b"barshoom")
            tree.add(["adir/afile"], ["file"])
            self.assertTrue(tree.is_versioned("adir/afile"))
            self.assertTrue(tree.is_versioned("adir"))

    def test_commit_trivial(self):
        """Smoke test for commit on a MemoryTree.

        Becamse of commits design and layering, if this works, all commit
        logic should work quite reliably.
        """
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(["", "foo"], kinds=["directory", "file"])
            tree.put_file_bytes_non_atomic("foo", b"barshoom")
            revision_id = tree.commit("message baby")
            # the parents list for the tree should have changed.
            self.assertEqual([revision_id], tree.get_parent_ids())
        # and we should have a revision that is accessible outside the tree lock
        revtree = tree.branch.repository.revision_tree(revision_id)
        with revtree.lock_read():
            self.assertEqual(b"barshoom", revtree.get_file("foo").read())

    def test_unversion(self):
        """Some test for unversion of a memory tree."""
        branch = self.make_branch("branch")
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(["", "foo"], kinds=["directory", "file"])
            tree.unversion(["foo"])
            self.assertFalse(tree.is_versioned("foo"))
            self.assertFalse(tree.has_filename("foo"))

    def test_last_revision(self):
        """There should be a last revision method we can call."""
        tree = self.make_branch_and_memory_tree("branch")
        with tree.lock_write():
            tree.add("")
            rev_id = tree.commit("first post")
        self.assertEqual(rev_id, tree.last_revision())

    def test_rename_file(self):
        tree = self.make_branch_and_memory_tree("branch")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add(["", "foo"], ["directory", "file"], ids=[b"root-id", b"foo-id"])
        tree.put_file_bytes_non_atomic("foo", b"content\n")
        tree.commit("one", rev_id=b"rev-one")
        tree.rename_one("foo", "bar")
        self.assertEqual("bar", tree.id2path(b"foo-id"))
        self.assertEqual(b"content\n", tree._file_transport.get_bytes("bar"))
        self.assertRaises(NoSuchFile, tree._file_transport.get_bytes, "foo")
        tree.commit("two", rev_id=b"rev-two")
        self.assertEqual(b"content\n", tree._file_transport.get_bytes("bar"))
        self.assertRaises(NoSuchFile, tree._file_transport.get_bytes, "foo")

        rev_tree2 = tree.branch.repository.revision_tree(b"rev-two")
        self.assertEqual("bar", rev_tree2.id2path(b"foo-id"))
        self.assertEqual(b"content\n", rev_tree2.get_file_text("bar"))

    def test_rename_file_to_subdir(self):
        tree = self.make_branch_and_memory_tree("branch")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add("")
        tree.mkdir("subdir", b"subdir-id")
        tree.add("foo", "file", b"foo-id")
        tree.put_file_bytes_non_atomic("foo", b"content\n")
        tree.commit("one", rev_id=b"rev-one")

        tree.rename_one("foo", "subdir/bar")
        self.assertEqual("subdir/bar", tree.id2path(b"foo-id"))
        self.assertEqual(b"content\n", tree._file_transport.get_bytes("subdir/bar"))
        tree.commit("two", rev_id=b"rev-two")
        rev_tree2 = tree.branch.repository.revision_tree(b"rev-two")
        self.assertEqual("subdir/bar", rev_tree2.id2path(b"foo-id"))



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Tree and InterTree."""

from breezy import errors, revision
from breezy.tests import TestCase, TestCaseWithTransport
from breezy.tree import (
    FileTimestampUnavailable,
    InterTree,
    find_previous_paths,
    get_canonical_path,
)


class TestErrors(TestCase):
    def test_file_timestamp_unavailable(self):
        e = FileTimestampUnavailable("/path/foo")
        self.assertEqual("The filestamp for /path/foo is not available.", str(e))


class TestInterTree(TestCaseWithTransport):
    def test_revision_tree_revision_tree(self):
        # we should have an InterTree registered for RevisionTree to
        # RevisionTree.
        tree = self.make_branch_and_tree(".")
        rev_id = tree.commit("first post")
        rev_id2 = tree.commit("second post", allow_pointless=True)
        rev_tree = tree.branch.repository.revision_tree(rev_id)
        rev_tree2 = tree.branch.repository.revision_tree(rev_id2)
        optimiser = InterTree.get(rev_tree, rev_tree2)
        self.assertIsInstance(optimiser, InterTree)
        optimiser = InterTree.get(rev_tree2, rev_tree)
        self.assertIsInstance(optimiser, InterTree)

    def test_working_tree_revision_tree(self):
        # we should have an InterTree available for WorkingTree to
        # RevisionTree.
        tree = self.make_branch_and_tree(".")
        rev_id = tree.commit("first post")
        rev_tree = tree.branch.repository.revision_tree(rev_id)
        optimiser = InterTree.get(rev_tree, tree)
        self.assertIsInstance(optimiser, InterTree)
        optimiser = InterTree.get(tree, rev_tree)
        self.assertIsInstance(optimiser, InterTree)

    def test_working_tree_working_tree(self):
        # we should have an InterTree available for WorkingTree to
        # WorkingTree.
        tree = self.make_branch_and_tree("1")
        tree2 = self.make_branch_and_tree("2")
        optimiser = InterTree.get(tree, tree2)
        self.assertIsInstance(optimiser, InterTree)
        optimiser = InterTree.get(tree2, tree)
        self.assertIsInstance(optimiser, InterTree)


class RecordingOptimiser(InterTree):
    calls: list[tuple[str, ...]] = []

    def compare(
        self,
        want_unchanged=False,
        specific_files=None,
        extra_trees=None,
        require_versioned=False,
        include_root=False,
        want_unversioned=False,
    ):
        self.calls.append(
            (
                "compare",
                self.source,
                self.target,
                want_unchanged,
                specific_files,
                extra_trees,
                require_versioned,
                include_root,
                want_unversioned,
            )
        )

    def find_source_path(self, target_path, recurse="none"):
        self.calls.append(
            ("find_source_path", self.source, self.target, target_path, recurse)
        )

    @classmethod
    def is_compatible(klass, source, target):
        return True


class TestTree(TestCaseWithTransport):
    def test_compare_calls_InterTree_compare(self):
        """This test tests the way Tree.compare() uses InterTree."""
        old_optimisers = InterTree._optimisers
        try:
            InterTree._optimisers = []
            RecordingOptimiser.calls = []
            InterTree.register_optimiser(RecordingOptimiser)
            tree = self.make_branch_and_tree("1")
            null_tree = tree.basis_tree()
            tree2 = self.make_branch_and_tree("2")
            # do a series of calls:
            # trivial usage
            tree.changes_from(tree2)
            # pass in all optional arguments by position
            tree.changes_from(tree2, "unchanged", "specific", "extra", "require", True)
            # pass in all optional arguments by keyword
            tree.changes_from(
                tree2,
                specific_files="specific",
                want_unchanged="unchanged",
                extra_trees="extra",
                require_versioned="require",
                include_root=True,
                want_unversioned=True,
            )
        finally:
            InterTree._optimisers = old_optimisers
        self.assertEqual(
            [
                ("find_source_path", null_tree, tree, "", "none"),
                ("find_source_path", null_tree, tree2, "", "none"),
                ("compare", tree2, tree, False, None, None, False, False, False),
                (
                    "compare",
                    tree2,
                    tree,
                    "unchanged",
                    "specific",
                    "extra",
                    "require",
                    True,
                    False,
                ),
                (
                    "compare",
                    tree2,
                    tree,
                    "unchanged",
                    "specific",
                    "extra",
                    "require",
                    True,
                    True,
                ),
            ],
            RecordingOptimiser.calls,
        )

    def test_changes_from_with_root(self):
        """Ensure the include_root option does what's expected."""
        wt = self.make_branch_and_tree(".")
        delta = wt.changes_from(wt.basis_tree())
        self.assertEqual(len(delta.added), 0)
        delta = wt.changes_from(wt.basis_tree(), include_root=True)
        self.assertEqual(len(delta.added), 1)
        self.assertEqual(delta.added[0].path[1], "")

    def test_changes_from_with_require_versioned(self):
        """Ensure the require_versioned option does what's expected."""
        wt = self.make_branch_and_tree(".")
        self.build_tree(["known_file", "unknown_file"])
        wt.add("known_file")

        self.assertRaises(
            errors.PathsNotVersionedError,
            wt.changes_from,
            wt.basis_tree(),
            wt,
            specific_files=["known_file", "unknown_file"],
            require_versioned=True,
        )

        # we need to pass a known file with an unknown file to get this to
        # fail when expected.
        delta = wt.changes_from(
            wt.basis_tree(),
            specific_files=["known_file", "unknown_file"],
            require_versioned=False,
        )
        self.assertEqual(len(delta.added), 1)


class FindPreviousPathsTests(TestCaseWithTransport):
    def test_new(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/b"])
        tree.add(["b"])
        revid1 = tree.commit("first")
        tree1 = tree.branch.repository.revision_tree(revid1)

        tree0 = tree.branch.repository.revision_tree(revision.NULL_REVISION)

        self.assertEqual({"b": None}, find_previous_paths(tree1, tree0, ["b"]))

    def test_find_previous_paths(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/b"])
        tree.add(["b"])
        revid1 = tree.commit("first")
        tree1 = tree.branch.repository.revision_tree(revid1)

        tree.rename_one("b", "c")
        self.build_tree(["tree/b"])
        tree.add(["b"])
        revid2 = tree.commit("second")
        tree2 = tree.branch.repository.revision_tree(revid2)

        self.assertEqual(
            {"c": "b", "b": None}, find_previous_paths(tree2, tree1, ["b", "c"])
        )


class GetCanonicalPath(TestCaseWithTransport):
    def test_existing_case(self):
        # Test that we can find a file from a path with different case
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/b"])
        tree.add(["b"])
        self.assertEqual("b", get_canonical_path(tree, "b", lambda x: x.lower()))
        self.assertEqual("b", get_canonical_path(tree, "B", lambda x: x.lower()))

    def test_nonexistant_preserves_case(self):
        tree = self.make_branch_and_tree("tree")
        self.assertEqual("b", get_canonical_path(tree, "b", lambda x: x.lower()))
        self.assertEqual("B", get_canonical_path(tree, "B", lambda x: x.lower()))

    def test_in_directory_with_case(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a/", "tree/a/b"])
        tree.add(["a", "a/b"])
        self.assertEqual("a/b", get_canonical_path(tree, "a/b", lambda x: x.lower()))
        self.assertEqual("a/b", get_canonical_path(tree, "A/B", lambda x: x.lower()))
        self.assertEqual("a/b", get_canonical_path(tree, "A/b", lambda x: x.lower()))
        self.assertEqual("a/C", get_canonical_path(tree, "A/C", lambda x: x.lower()))

    def test_trailing_slash(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a/", "tree/a/b"])
        tree.add(["a", "a/b"])
        self.assertEqual("a", get_canonical_path(tree, "a", lambda x: x))
        self.assertEqual("a", get_canonical_path(tree, "a/", lambda x: x))



#    __init__.py -- Testsuite for quilt
#    Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
#    Breezy is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Breezy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Breezy; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

__all__ = [
    "TestCaseInTempDir",
    "TestCaseWithTransport",
    "TestUtil",
    "load_tests",
    "multiply_tests",
    "quilt_feature",
]


from ....tests import TestCaseInTempDir, TestCaseWithTransport, TestUtil, multiply_tests
from ....tests.features import ExecutableFeature

quilt_feature = ExecutableFeature("quilt")


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
        "test_merge",
        "test_wrapper",
    ]
    basic_tests.addTest(
        loader.loadTestsFromModuleNames(
            ["{}.{}".format(__name__, i) for i in testmod_names]
        )
    )

    return basic_tests



# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for the switch command of bzr."""

import os

from breezy import branch, osutils, urlutils
from breezy.controldir import ControlDir
from breezy.directory_service import directories
from breezy.tests import TestCaseWithTransport, script
from breezy.tests.features import UnicodeFilenameFeature
from breezy.workingtree import WorkingTree


class TestSwitch(TestCaseWithTransport):
    def _create_sample_tree(self):
        tree = self.make_branch_and_tree("branch-1")
        self.build_tree(["branch-1/file-1", "branch-1/file-2"])
        tree.add("file-1")
        tree.commit("rev1")
        tree.add("file-2")
        tree.commit("rev2")
        return tree

    def test_switch_up_to_date_light_checkout(self):
        self.make_branch_and_tree("branch")
        self.run_bzr("branch branch branch2")
        self.run_bzr("checkout --lightweight branch checkout")
        os.chdir("checkout")
        out, err = self.run_bzr("switch ../branch2")
        self.assertContainsRe(err, "Tree is up to date at revision 0.\n")
        self.assertContainsRe(err, "Switched to branch at .*/branch2.\n")
        self.assertEqual("", out)

    def test_switch_out_of_date_light_checkout(self):
        self.make_branch_and_tree("branch")
        self.run_bzr("branch branch branch2")
        self.build_tree(["branch2/file"])
        self.run_bzr("add branch2/file")
        self.run_bzr("commit -m add-file branch2")
        self.run_bzr("checkout --lightweight branch checkout")
        os.chdir("checkout")
        out, err = self.run_bzr("switch ../branch2")
        # self.assertContainsRe(err, '\+N  file')
        self.assertContainsRe(err, "Updated to revision 1.\n")
        self.assertContainsRe(err, "Switched to branch at .*/branch2.\n")
        self.assertEqual("", out)

    def _test_switch_nick(self, lightweight):
        """Check that the nick gets switched too."""
        tree1 = self.make_branch_and_tree("branch1")
        tree2 = self.make_branch_and_tree("branch2")
        tree2.pull(tree1.branch)
        checkout = tree1.branch.create_checkout("checkout", lightweight=lightweight)
        self.assertEqual(checkout.branch.nick, tree1.branch.nick)
        self.assertEqual(checkout.branch.get_config().has_explicit_nickname(), False)
        self.run_bzr("switch branch2", working_dir="checkout")

        # we need to get the tree again, otherwise we don't get the new branch
        checkout = WorkingTree.open("checkout")
        self.assertEqual(checkout.branch.nick, tree2.branch.nick)
        self.assertEqual(checkout.branch.get_config().has_explicit_nickname(), False)

    def test_switch_nick(self):
        self._test_switch_nick(lightweight=False)

    def test_switch_nick_lightweight(self):
        self._test_switch_nick(lightweight=True)

    def _test_switch_explicit_nick(self, lightweight):
        """Check that the nick gets switched too."""
        tree1 = self.make_branch_and_tree("branch1")
        tree2 = self.make_branch_and_tree("branch2")
        tree2.pull(tree1.branch)
        checkout = tree1.branch.create_checkout("checkout", lightweight=lightweight)
        self.assertEqual(checkout.branch.nick, tree1.branch.nick)
        checkout.branch.nick = "explicit_nick"
        self.assertEqual(checkout.branch.nick, "explicit_nick")
        self.assertEqual(
            checkout.branch.get_config()._get_explicit_nickname(), "explicit_nick"
        )
        self.run_bzr("switch branch2", working_dir="checkout")

        # we need to get the tree again, otherwise we don't get the new branch
        checkout = WorkingTree.open("checkout")
        self.assertEqual(checkout.branch.nick, tree2.branch.nick)
        self.assertEqual(
            checkout.branch.get_config()._get_explicit_nickname(), tree2.branch.nick
        )

    def test_switch_explicit_nick(self):
        self._test_switch_explicit_nick(lightweight=False)

    def test_switch_explicit_nick_lightweight(self):
        self._test_switch_explicit_nick(lightweight=True)

    def test_switch_finds_relative_branch(self):
        """Switch will find 'foo' relative to the branch the checkout is of."""
        self.build_tree(["repo/"])
        tree1 = self.make_branch_and_tree("repo/brancha")
        tree1.commit("foo")
        tree2 = self.make_branch_and_tree("repo/branchb")
        tree2.pull(tree1.branch)
        branchb_id = tree2.commit("bar")
        checkout = tree1.branch.create_checkout("checkout", lightweight=True)
        self.run_bzr(["switch", "branchb"], working_dir="checkout")
        self.assertEqual(branchb_id, checkout.last_revision())
        checkout = checkout.controldir.open_workingtree()
        self.assertEqual(tree2.branch.base, checkout.branch.base)

    def test_switch_finds_relative_bound_branch(self):
        """Using switch on a heavy checkout should find master sibling.

        The behaviour of lighweight and heavy checkouts should be
        consistent when using the convenient "switch to sibling" feature
        Both should switch to a sibling of the branch
        they are bound to, and not a sibling of themself
        """
        self.build_tree(["repo/", "heavyco/"])
        tree1 = self.make_branch_and_tree("repo/brancha")
        tree1.commit("foo")
        tree2 = self.make_branch_and_tree("repo/branchb")
        tree2.pull(tree1.branch)
        branchb_id = tree2.commit("bar")
        checkout = tree1.branch.create_checkout("heavyco/a", lightweight=False)
        self.run_bzr(["switch", "branchb"], working_dir="heavyco/a")
        # Refresh checkout as 'switch' modified it
        checkout = checkout.controldir.open_workingtree()
        self.assertEqual(branchb_id, checkout.last_revision())
        self.assertEqual(tree2.branch.base, checkout.branch.get_bound_location())

    def test_switch_finds_relative_unicode_branch(self):
        """Switch will find 'foo' relative to the branch the checkout is of."""
        self.requireFeature(UnicodeFilenameFeature)
        self.build_tree(["repo/"])
        tree1 = self.make_branch_and_tree("repo/brancha")
        tree1.commit("foo")
        tree2 = self.make_branch_and_tree("repo/branch\xe9")
        tree2.pull(tree1.branch)
        branchb_id = tree2.commit("bar")
        checkout = tree1.branch.create_checkout("checkout", lightweight=True)
        self.run_bzr(["switch", "branch\xe9"], working_dir="checkout")
        self.assertEqual(branchb_id, checkout.last_revision())
        checkout = checkout.controldir.open_workingtree()
        self.assertEqual(tree2.branch.base, checkout.branch.base)

    def test_switch_revision(self):
        tree = self._create_sample_tree()
        tree.branch.create_checkout("checkout", lightweight=True)
        self.run_bzr(["switch", "branch-1", "-r1"], working_dir="checkout")
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")

    def test_switch_into_colocated(self):
        # Create a new colocated branch from an existing non-colocated branch.
        tree = self.make_branch_and_tree(".", format="development-colo")
        self.build_tree(["file-1", "file-2"])
        tree.add("file-1")
        tree.commit("rev1")
        tree.add("file-2")
        tree.commit("rev2")
        self.run_bzr(["switch", "-b", "anotherbranch"])
        self.assertEqual(
            {"", "anotherbranch"}, set(tree.branch.controldir.branch_names())
        )

    def test_switch_into_unrelated_colocated(self):
        # Create a new colocated branch from an existing non-colocated branch.
        tree = self.make_branch_and_tree(".", format="development-colo")
        self.build_tree(["file-1", "file-2"])
        tree.add("file-1")
        tree.commit("rev1")
        tree.add("file-2")
        tree.commit("rev2")
        tree.controldir.create_branch(name="foo")
        self.run_bzr_error(["Cannot switch a branch, only a checkout."], "switch foo")
        self.run_bzr(["switch", "--force", "foo"])

    def test_switch_existing_colocated(self):
        # Create a branch branch-1 that initially is a checkout of 'foo'
        # Use switch to change it to 'anotherbranch'
        repo = self.make_repository("branch-1", format="development-colo")
        target_branch = repo.controldir.create_branch(name="foo")
        repo.controldir.set_branch_reference(target_branch)
        tree = repo.controldir.create_workingtree()
        self.build_tree(["branch-1/file-1", "branch-1/file-2"])
        tree.add("file-1")
        revid1 = tree.commit("rev1")
        tree.add("file-2")
        tree.commit("rev2")
        otherbranch = tree.controldir.create_branch(name="anotherbranch")
        otherbranch.generate_revision_history(revid1)
        self.run_bzr(["switch", "anotherbranch"], working_dir="branch-1")
        tree = WorkingTree.open("branch-1")
        self.assertEqual(tree.last_revision(), revid1)
        self.assertEqual(tree.branch.control_url, otherbranch.control_url)

    def test_switch_new_colocated(self):
        # Create a branch branch-1 that initially is a checkout of 'foo'
        # Use switch to create 'anotherbranch' which derives from that
        repo = self.make_repository("branch-1", format="development-colo")
        target_branch = repo.controldir.create_branch(name="foo")
        repo.controldir.set_branch_reference(target_branch)
        tree = repo.controldir.create_workingtree()
        self.build_tree(["branch-1/file-1", "branch-1/file-2"])
        tree.add("file-1")
        revid1 = tree.commit("rev1")
        self.run_bzr(["switch", "-b", "anotherbranch"], working_dir="branch-1")
        bzrdir = ControlDir.open("branch-1")
        self.assertEqual(
            {b.name for b in bzrdir.list_branches()}, {"foo", "anotherbranch"}
        )
        self.assertEqual(bzrdir.open_branch().name, "anotherbranch")
        self.assertEqual(bzrdir.open_branch().last_revision(), revid1)

    def test_switch_new_colocated_unicode(self):
        # Create a branch branch-1 that initially is a checkout of 'foo'
        # Use switch to create 'branch\xe9' which derives from that
        self.requireFeature(UnicodeFilenameFeature)
        repo = self.make_repository("branch-1", format="development-colo")
        target_branch = repo.controldir.create_branch(name="foo")
        repo.controldir.set_branch_reference(target_branch)
        tree = repo.controldir.create_workingtree()
        self.build_tree(["branch-1/file-1", "branch-1/file-2"])
        tree.add("file-1")
        revid1 = tree.commit("rev1")
        self.run_bzr(["switch", "-b", "branch\xe9"], working_dir="branch-1")
        bzrdir = ControlDir.open("branch-1")
        self.assertEqual(
            {b.name for b in bzrdir.list_branches()}, {"foo", "branch\xe9"}
        )
        self.assertEqual(bzrdir.open_branch().name, "branch\xe9")
        self.assertEqual(bzrdir.open_branch().last_revision(), revid1)

    def test_switch_only_revision(self):
        tree = self._create_sample_tree()
        tree.branch.create_checkout("checkout", lightweight=True)
        self.assertPathExists("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        self.run_bzr(["switch", "-r1"], working_dir="checkout")
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")
        # Check that we don't accept a range
        self.run_bzr_error(
            ["brz switch --revision takes exactly one revision identifier"],
            ["switch", "-r0..2"],
            working_dir="checkout",
        )

    def prepare_lightweight_switch(self):
        branch = self.make_branch("branch")
        branch.create_checkout("tree", lightweight=True)
        osutils.rename("branch", "branch1")

    def test_switch_lightweight_after_branch_moved(self):
        self.prepare_lightweight_switch()
        self.run_bzr("switch --force ../branch1", working_dir="tree")
        branch_location = WorkingTree.open("tree").branch.base
        self.assertEndsWith(branch_location, "branch1/")

    def test_switch_lightweight_after_branch_moved_relative(self):
        self.prepare_lightweight_switch()
        self.run_bzr("switch --force branch1", working_dir="tree")
        branch_location = WorkingTree.open("tree").branch.base
        self.assertEndsWith(branch_location, "branch1/")

    def test_create_branch_no_branch(self):
        self.prepare_lightweight_switch()
        self.run_bzr_error(
            ["cannot create branch without source branch"],
            "switch --create-branch ../branch2",
            working_dir="tree",
        )

    def test_create_branch(self):
        branch = self.make_branch("branch")
        tree = branch.create_checkout("tree", lightweight=True)
        tree.commit("one", rev_id=b"rev-1")
        self.run_bzr("switch --create-branch ../branch2", working_dir="tree")
        tree = WorkingTree.open("tree")
        self.assertEndsWith(tree.branch.base, "/branch2/")

    def test_create_branch_local(self):
        branch = self.make_branch("branch")
        tree = branch.create_checkout("tree", lightweight=True)
        tree.commit("one", rev_id=b"rev-1")
        self.run_bzr("switch --create-branch branch2", working_dir="tree")
        tree = WorkingTree.open("tree")
        # The new branch should have been created at the same level as
        # 'branch', because we did not have a '/' segment
        self.assertEqual(branch.base[:-1] + "2/", tree.branch.base)

    def test_create_branch_short_name(self):
        branch = self.make_branch("branch")
        tree = branch.create_checkout("tree", lightweight=True)
        tree.commit("one", rev_id=b"rev-1")
        self.run_bzr("switch -b branch2", working_dir="tree")
        tree = WorkingTree.open("tree")
        # The new branch should have been created at the same level as
        # 'branch', because we did not have a '/' segment
        self.assertEqual(branch.base[:-1] + "2/", tree.branch.base)

    def test_create_branch_directory_services(self):
        branch = self.make_branch("branch")
        tree = branch.create_checkout("tree", lightweight=True)

        class FooLookup:
            def look_up(self, name, url, purpose=None):
                return "foo-" + name

        directories.register("foo:", FooLookup, "Create branches named foo-")
        self.addCleanup(directories.remove, "foo:")
        self.run_bzr("switch -b foo:branch2", working_dir="tree")
        tree = WorkingTree.open("tree")
        self.assertEndsWith(tree.branch.base, "foo-branch2/")

    def test_switch_with_post_switch_hook(self):
        from breezy import branch as _mod_branch

        calls = []
        _mod_branch.Branch.hooks.install_named_hook("post_switch", calls.append, None)
        self.make_branch_and_tree("branch")
        self.run_bzr("branch branch branch2")
        self.run_bzr("checkout branch checkout")
        os.chdir("checkout")
        self.assertLength(0, calls)
        out, err = self.run_bzr("switch ../branch2")
        self.assertLength(1, calls)

    def test_switch_lightweight_co_with_post_switch_hook(self):
        from breezy import branch as _mod_branch

        calls = []
        _mod_branch.Branch.hooks.install_named_hook("post_switch", calls.append, None)
        self.make_branch_and_tree("branch")
        self.run_bzr("branch branch branch2")
        self.run_bzr("checkout --lightweight branch checkout")
        os.chdir("checkout")
        self.assertLength(0, calls)
        out, err = self.run_bzr("switch ../branch2")
        self.assertLength(1, calls)

    def test_switch_lightweight_directory(self):
        """Test --directory option."""
        # create a source branch
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="initial")

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("b/a", b"initial\nmore\n")])
        b_tree.commit(message="more")

        self.run_bzr("checkout --lightweight a checkout")
        self.run_bzr("switch --directory checkout b")
        self.assertFileEqual(b"initial\nmore\n", "checkout/a")


class TestSwitchParentLocationBase(TestCaseWithTransport):
    def setUp(self):
        """Set up a repository and branch ready for testing."""
        super().setUp()
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
                $ brz init-shared-repo --no-trees repo
                Shared repository...
                Location:
                  shared repository: repo
                $ brz init repo/trunk
                Created a repository branch...
                Using shared repository: ...
                """,
        )

    def assertParent(self, expected_parent, branch):
        """Verify that the parent is not None and is set correctly."""
        branch.get_parent()
        self.assertIsSameRealPath(
            urlutils.local_path_to_url(expected_parent), branch.get_parent()
        )


class TestSwitchParentLocation(TestSwitchParentLocationBase):
    def _checkout_and_switch(self, option=""):
        self.script_runner.run_script(
            self,
            """
                $ brz checkout {option} repo/trunk checkout
                $ cd checkout
                $ brz switch --create-branch switched
                2>Tree is up to date at revision 0.
                2>Switched to branch at .../switched/
                $ cd ..
                """.format(**locals()),
        )
        bound_branch = branch.Branch.open_containing("checkout")[0]
        master_branch = branch.Branch.open_containing("repo/switched")[0]
        return (bound_branch, master_branch)

    def test_switch_parent_lightweight(self):
        """Lightweight checkout using brz switch."""
        bb, mb = self._checkout_and_switch(option="--lightweight")
        self.assertParent("repo/trunk", bb)
        self.assertParent("repo/trunk", mb)

    def test_switch_parent_heavyweight(self):
        """Heavyweight checkout using brz switch."""
        bb, mb = self._checkout_and_switch()
        self.assertParent("repo/trunk", bb)
        self.assertParent("repo/trunk", mb)


class TestSwitchDoesntOpenMasterBranch(TestCaseWithTransport):
    # See https://bugs.launchpad.net/bzr/+bug/812285
    # "brz switch --create-branch" can point the new branch's parent to the
    # master branch, but it doesn't have to open it to do so.

    def test_switch_create_doesnt_open_master_branch(self):
        master = self.make_branch_and_tree("master")
        master.commit("one")
        # Note: not a lightweight checkout
        master.branch.create_checkout("checkout")
        opened = []

        def open_hook(branch):
            # Just append the final directory of the branch
            name = branch.base.rstrip("/").rsplit("/", 1)[1]
            opened.append(name)

        branch.Branch.hooks.install_named_hook("open", open_hook, "open_hook_logger")
        self.run_bzr("switch --create-branch -d checkout feature")
        # We only open the master branch 1 time.
        # This test should be cleaner to write, but see bug:
        #  https://bugs.launchpad.net/bzr/+bug/812295
        self.assertEqual(1, opened.count("master"))


class TestSwitchUncommitted(TestCaseWithTransport):
    def prepare(self):
        tree = self.make_branch_and_tree("orig")
        tree.commit("")
        tree.branch.controldir.sprout("new")
        checkout = tree.branch.create_checkout("checkout", lightweight=True)
        self.build_tree(["checkout/a"])
        self.assertPathExists("checkout/a")
        checkout.add("a")
        return checkout

    def test_store_and_restore_uncommitted(self):
        checkout = self.prepare()
        self.run_bzr(["switch", "--store", "-d", "checkout", "new"])
        self.build_tree(["checkout/b"])
        checkout.add("b")
        self.assertPathDoesNotExist("checkout/a")
        self.assertPathExists("checkout/b")
        self.run_bzr(["switch", "--store", "-d", "checkout", "orig"])
        self.assertPathExists("checkout/a")
        self.assertPathDoesNotExist("checkout/b")

    def test_does_not_store(self):
        self.prepare()
        self.run_bzr(["switch", "-d", "checkout", "new"])
        self.assertPathExists("checkout/a")

    def test_does_not_restore_changes(self):
        self.prepare()
        self.run_bzr(["switch", "--store", "-d", "checkout", "new"])
        self.assertPathDoesNotExist("checkout/a")
        self.run_bzr(["switch", "-d", "checkout", "orig"])
        self.assertPathDoesNotExist("checkout/a")


class TestSwitchStandAloneCorruption(TestCaseWithTransport):
    def test_empty_tree_switch(self):
        """Switch . on an empty tree gets infinite recursion.

        Inspired by: https://bugs.launchpad.net/bzr/+bug/1018628
        """
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
            $ brz init
            Created a standalone tree (format: 2a)
            $ brz switch .
            2>brz: ERROR: switching would create a branch reference loop. Use the "bzr up" command to switch to a different revision.
            """,
        )

    def test_switch_on_previous_rev(self):
        """Switch to previous rev in a standalone directory.

        Inspired by: https://bugs.launchpad.net/brz/+bug/1018628
        """
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
           $ brz init
           Created a standalone tree (format: 2a)
           $ brz commit -m 1 --unchanged
           $ brz commit -m 2 --unchanged
           $ brz switch -r 1
           2>brz: ERROR: switching would create a branch reference loop. Use the "bzr up" command to switch to a different revision.""",
            null_output_matches_anything=True,
        )

    def test_switch_create_colo_locks_repo_path(self):
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
            $ mkdir mywork
            $ cd mywork
            $ brz init
            Created a standalone tree (format: 2a)
            $ echo A > a && brz add a && brz commit -m A
            $ brz switch -b br1
            $ cd ..
            $ mv mywork mywork1
            $ cd mywork1
            $ brz branches
              br1
            """,
            null_output_matches_anything=True,
        )

    def test_switch_to_new_branch_on_old_rev(self):
        """Switch to previous rev in a standalone directory.

        Inspired by: https://bugs.launchpad.net/brz/+bug/933362
        """
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
           $ brz init
           Created a standalone tree (format: 2a)
           $ brz switch -b trunk
           2>Tree is up to date at revision 0.
           2>Switched to branch trunk
           $ brz commit -m 1 --unchanged
           2>Committing to: ...
           2>Committed revision 1.
           $ brz commit -m 2 --unchanged
           2>Committing to: ...
           2>Committed revision 2.
           $ brz switch -b blah -r1
           2>Updated to revision 1.
           2>Switched to branch blah
           $ brz branches
           * blah
             trunk
           $ brz st
           """,
        )



"""apport package hook for Breezy."""

# Copyright (c) 2009, 2010 Canonical Ltd.
# Author: Matt Zimmerman <mdz@canonical.com>
#         and others

import os

from apport.hookutils import *  # noqa: F403

brz_log = os.path.expanduser("~/.brz.log")
dot_brz = os.path.expanduser("~/.config/breezy")


def _add_log_tail(report):
    # may have already been added in-process
    if "BrzLogTail" in report:
        return

    with open(brz_log) as f:
        brz_log_lines = f.readlines()
    brz_log_lines.reverse()

    brz_log_tail = []
    blanks = 0
    for line in brz_log_lines:
        if line == "\n":
            blanks += 1
        brz_log_tail.append(line)
        if blanks >= 2:
            break

    brz_log_tail.reverse()
    report["BrzLogTail"] = "".join(brz_log_tail)


def add_info(report):
    _add_log_tail(report)
    if "BrzPlugins" not in report:
        # may already be present in-process
        report["BrzPlugins"] = command_output(["brz", "plugins", "-v"])

    # by default assume brz crashes are upstream bugs; this relies on
    # having a brz entry under /etc/apport/crashdb.conf.d/
    report["CrashDB"] = "brz"

    # these may contain some sensitive info (smtp_passwords)
    # TODO: strip that out and attach the rest

    # attach_file_if_exists(report,
    # os.path.join(dot_brz, 'breezy.conf', 'BrzConfig')
    # attach_file_if_exists(report,
    # os.path.join(dot_brz, 'locations.conf', 'BrzLocations')


# vim: expandtab shiftwidth=4



# Copyright (C) 2006-2012 Aaron Bentley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the 'brz import' command."""

import os

from ... import tests
from .. import features

LzmaFeature = features.ModuleAvailableFeature("lzma")


class TestImport(tests.TestCaseWithTransport):
    def test_import_upstream(self):
        self.run_bzr("init source")
        os.mkdir("source/src")
        with open("source/src/myfile", "wb") as f:
            f.write(b"hello?")
        os.chdir("source")
        self.run_bzr("add")
        self.run_bzr("commit -m hello")
        self.run_bzr("export ../source-0.1.tar.gz")
        self.run_bzr("export ../source-0.1.tar.bz2")
        self.run_bzr("export ../source-0.1")
        self.run_bzr("init ../import")
        os.chdir("../import")
        self.run_bzr("import ../source-0.1.tar.gz")
        self.assertPathExists("src/myfile")
        result = self.run_bzr("import ../source-0.1.tar.gz", retcode=3)[1]
        self.assertContainsRe(result, "Working tree has uncommitted changes")
        self.run_bzr("commit -m commit")
        self.run_bzr("import ../source-0.1.tar.gz")
        os.chdir("..")
        self.run_bzr("init import2")
        self.run_bzr("import source-0.1.tar.gz import2")
        self.assertPathExists("import2/src/myfile")
        self.run_bzr("import source-0.1.tar.gz import3")
        self.assertPathExists("import3/src/myfile")
        self.run_bzr("import source-0.1.tar.bz2 import4")
        self.assertPathExists("import4/src/myfile")
        self.run_bzr("import source-0.1 import5")
        self.assertPathExists("import5/src/myfile")

    def test_import_upstream_lzma(self):
        self.requireFeature(LzmaFeature)
        self.run_bzr("init source")
        os.mkdir("source/src")
        with open("source/src/myfile", "wb") as f:
            f.write(b"hello?")
        os.chdir("source")
        self.run_bzr("add")
        self.run_bzr("commit -m hello")
        self.run_bzr("export ../source-0.1.tar.lzma")
        self.run_bzr("export ../source-0.1.tar.xz")
        os.chdir("..")
        self.run_bzr("import source-0.1.tar.lzma import1")
        self.assertPathExists("import1/src/myfile")
        self.run_bzr("import source-0.1.tar.xz import2")
        self.assertPathExists("import2/src/myfile")



#    __init__.py -- Quilt support for breezy
#
#    Breezy is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Breezy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Breezy; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Quilt patch integration.

This plugin adds support for three configuration options:

 * quilt.commit_policy
 * quilt.smart_merge
 * quilt.tree_policy

"""

from ... import trace
from ...errors import BzrError


class QuiltUnapplyError(BzrError):
    _fmt = "Unable to unapply quilt patches for %(kind)r tree: %(msg)s"

    def __init__(self, kind, msg):
        BzrError.__init__(self)
        self.kind = kind
        if msg is not None and msg.count("\n") == 1:
            msg = msg.strip()
        self.msg = msg


def pre_merge_quilt(merger):
    if getattr(merger, "_no_quilt_unapplying", False):
        return

    config = merger.working_tree.get_config_stack()
    merger.debuild_config = config
    if not config.get("quilt.smart_merge"):
        trace.mutter("skipping smart quilt merge, not enabled.")
        return

    if (
        not merger.other_tree.is_versioned(".pc")
        and not merger.this_tree.is_versioned(".pc")
        and not merger.working_tree.is_versioned(".pc")
    ):
        return

    from .quilt import QuiltError, QuiltPatches

    quilt = QuiltPatches(merger.working_tree)
    from .merge import tree_unapply_patches

    trace.note("Unapplying quilt patches to prevent spurious conflicts")
    merger._quilt_tempdirs = []
    merger._old_quilt_series = quilt.series()
    if merger._old_quilt_series:
        quilt.pop_all()
    try:
        merger.this_tree, this_dir = tree_unapply_patches(
            merger.this_tree, merger.this_branch, force=True
        )
    except QuiltError as e:
        raise QuiltUnapplyError("this", e.stderr)
    else:
        if this_dir is not None:
            merger._quilt_tempdirs.append(this_dir)
    try:
        merger.base_tree, base_dir = tree_unapply_patches(
            merger.base_tree, merger.this_branch, force=True
        )
    except QuiltError as e:
        raise QuiltUnapplyError("base", e.stderr)
    else:
        if base_dir is not None:
            merger._quilt_tempdirs.append(base_dir)
    other_branch = getattr(merger, "other_branch", None)
    if other_branch is None:
        other_branch = merger.this_branch
    try:
        merger.other_tree, other_dir = tree_unapply_patches(
            merger.other_tree, other_branch, force=True
        )
    except QuiltError as e:
        raise QuiltUnapplyError("other", e.stderr)
    else:
        if other_dir is not None:
            merger._quilt_tempdirs.append(other_dir)


def post_merge_quilt_cleanup(merger):
    import shutil

    for dir in getattr(merger, "_quilt_tempdirs", []):
        shutil.rmtree(dir)
    config = merger.working_tree.get_config_stack()
    policy = config.get("quilt.tree_policy")
    if policy is None:
        return
    from .merge import post_process_quilt_patches

    post_process_quilt_patches(
        merger.working_tree, getattr(merger, "_old_quilt_series", []), policy
    )


def start_commit_check_quilt(tree):
    """start_commit hook which checks the state of quilt patches."""
    config = tree.get_config_stack()
    policy = config.get("quilt.commit_policy")
    from .merge import start_commit_quilt_patches
    from .wrapper import QuiltNotInstalled

    try:
        start_commit_quilt_patches(tree, policy)
    except QuiltNotInstalled:
        trace.warning("quilt not installed; not updating patches")


def post_build_tree_quilt(tree):
    config = tree.get_config_stack()
    policy = config.get("quilt.tree_policy")
    if policy is None:
        return
    from .merge import post_process_quilt_patches
    from .wrapper import QuiltNotInstalled

    try:
        post_process_quilt_patches(tree, [], policy)
    except QuiltNotInstalled:
        trace.warning("quilt not installed; not touching patches")


from ...hooks import install_lazy_named_hook

install_lazy_named_hook(
    "breezy.merge",
    "Merger.hooks",
    "pre_merge_quilt",
    pre_merge_quilt,
    "Quilt patch (un)applying",
)
install_lazy_named_hook(
    "breezy.mutabletree",
    "MutableTree.hooks",
    "start_commit",
    start_commit_check_quilt,
    "Check for (un)applied quilt patches",
)
install_lazy_named_hook(
    "breezy.merge",
    "Merger.hooks",
    "post_merge",
    post_merge_quilt_cleanup,
    "Cleaning up quilt temporary directories",
)
install_lazy_named_hook(
    "breezy.mutabletree",
    "MutableTree.hooks",
    "post_build_tree",
    post_build_tree_quilt,
    "Applying quilt patches.",
)


from ...config import Option, bool_from_store, option_registry

option_registry.register(
    Option(
        "quilt.smart_merge",
        default=True,
        from_unicode=bool_from_store,
        help="Unapply quilt patches before merging.",
    )
)


def policy_from_store(s):
    if s not in ("applied", "unapplied"):
        raise ValueError("Invalid quilt.commit_policy: {}".format(s))
    return s


option_registry.register(
    Option(
        "quilt.commit_policy",
        default=None,
        from_unicode=policy_from_store,
        help="Whether to apply or unapply all patches in commits.",
    )
)

option_registry.register(
    Option(
        "quilt.tree_policy",
        default=None,
        from_unicode=policy_from_store,
        help="Whether to apply or unapply all patches after checkout/update.",
    )
)


def load_tests(loader, basic_tests, pattern):
    basic_tests.addTest(loader.loadTestsFromModuleNames([__name__ + ".tests"]))
    return basic_tests
